import unittest
import os
import sys
sys.path.append(os.getcwd())
from tests import utils_for_test as test_utils
sys.path.pop()

# Execute these steps for standard tests:
# - Save this file as `test_{parser_name}.py` since the helper methods extract parser names from the filename.
# - Organize fixtures in `tests/fixtures` for optimal structure.
# - Format fixtures as follows (using double dashes):
#     - `{parser_name}--{some_test_description}.out` for command output.
#     - `{parser_name}--{some_test_description}.json` for expected JSON after parsing.

# create output with:
# curl -ILvs http://google.com 2> curl_head--ILvs-google-com.out
# cat curl_head--ILvs-google-com.out | jc --curl-head > curl_head--ILvs-google-com.json

class MyTests(unittest.TestCase):

    def test_curl_head_nodata(self):
        """
        Test 'curl_head' with no data
        """
        test_utils.run_no_data(self, __file__, [])

    def test_curl_head_all_fixtures(self):
        """
        Test 'curl_head' with various fixtures
        """
        test_utils.run_all_fixtures(self, __file__)


if __name__ == '__main__':
    unittest.main()



import os
import unittest
import json
import jc.parsers.traceroute_s

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class MyTests(unittest.TestCase):

    # input
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/traceroute.out'), 'r', encoding='utf-8') as f:
        centos_7_7_traceroute = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/traceroute-no-header.out'), 'r', encoding='utf-8') as f:
        osx_10_14_6_traceroute_noheader = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/traceroute-asn.out'), 'r', encoding='utf-8') as f:
        osx_10_14_6_traceroute_asn = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/traceroute-mult-addresses.out'), 'r', encoding='utf-8') as f:
        osx_10_14_6_traceroute_mult_addresses = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/traceroute-q.out'), 'r', encoding='utf-8') as f:
        osx_10_14_6_traceroute_q = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/traceroute.out'), 'r', encoding='utf-8') as f:
        osx_10_14_6_traceroute = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/traceroute6-mult-addresses.out'), 'r', encoding='utf-8') as f:
        osx_10_14_6_traceroute6_mult_addresses = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/traceroute6.out'), 'r', encoding='utf-8') as f:
        osx_10_14_6_traceroute6 = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/freebsd12/traceroute.out'), 'r', encoding='utf-8') as f:
        freebsd12_traceroute = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/freebsd12/traceroute6.out'), 'r', encoding='utf-8') as f:
        freebsd12_traceroute6 = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/traceroute1.out'), 'r', encoding='utf-8') as f:
        generic_traceroute1 = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/traceroute2.out'), 'r', encoding='utf-8') as f:
        generic_traceroute2 = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/traceroute3.out'), 'r', encoding='utf-8') as f:
        generic_traceroute3 = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/traceroute4.out'), 'r', encoding='utf-8') as f:
        generic_traceroute4 = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/traceroute5.out'), 'r', encoding='utf-8') as f:
        generic_traceroute5 = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/traceroute6.out'), 'r', encoding='utf-8') as f:
        generic_traceroute6 = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/traceroute7.out'), 'r', encoding='utf-8') as f:
        generic_traceroute7 = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/traceroute8.out'), 'r', encoding='utf-8') as f:
        generic_traceroute8 = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/traceroute-n-ipv4.out'), 'r', encoding='utf-8') as f:
        generic_traceroute_n_ipv4 = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/traceroute-n-q1-ipv4.out'), 'r', encoding='utf-8') as f:
        generic_traceroute_n_q1_ipv4 = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/traceroute-n-ipv6.out'), 'r', encoding='utf-8') as f:
        generic_traceroute_n_ipv6 = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/traceroute-long-ipv6.out'), 'r', encoding='utf-8') as f:
        generic_traceroute_long_ipv6 = f.read()


    # output
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/traceroute-no-header-streaming.json'), 'r', encoding='utf-8') as f:
        osx_10_14_6_traceroute_no_header_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/traceroute-streaming.json'), 'r', encoding='utf-8') as f:
        centos_7_7_traceroute_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/traceroute-asn-streaming.json'), 'r', encoding='utf-8') as f:
        osx_10_14_6_traceroute_asn_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/traceroute-mult-addresses-streaming.json'), 'r', encoding='utf-8') as f:
        osx_10_14_6_traceroute_mult_addresses_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/traceroute-q-streaming.json'), 'r', encoding='utf-8') as f:
        osx_10_14_6_traceroute_q_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/traceroute-streaming.json'), 'r', encoding='utf-8') as f:
        osx_10_14_6_traceroute_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/traceroute6-mult-addresses-streaming.json'), 'r', encoding='utf-8') as f:
        osx_10_14_6_traceroute6_mult_addresses_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/traceroute6-streaming.json'), 'r', encoding='utf-8') as f:
        osx_10_14_6_traceroute6_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/freebsd12/traceroute-streaming.json'), 'r', encoding='utf-8') as f:
        freebsd12_traceroute_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/freebsd12/traceroute6-streaming.json'), 'r', encoding='utf-8') as f:
        freebsd12_traceroute6_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/traceroute1-streaming.json'), 'r', encoding='utf-8') as f:
        generic_traceroute1_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/traceroute2-streaming.json'), 'r', encoding='utf-8') as f:
        generic_traceroute2_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/traceroute3-streaming.json'), 'r', encoding='utf-8') as f:
        generic_traceroute3_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/traceroute4-streaming.json'), 'r', encoding='utf-8') as f:
        generic_traceroute4_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/traceroute5-streaming.json'), 'r', encoding='utf-8') as f:
        generic_traceroute5_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/traceroute6-streaming.json'), 'r', encoding='utf-8') as f:
        generic_traceroute6_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/traceroute7-streaming.json'), 'r', encoding='utf-8') as f:
        generic_traceroute7_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/traceroute8-streaming.json'), 'r', encoding='utf-8') as f:
        generic_traceroute8_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/traceroute-n-ipv4-streaming.json'), 'r', encoding='utf-8') as f:
        generic_traceroute_n_ipv4_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/traceroute-n-q1-ipv4-streaming.json'), 'r', encoding='utf-8') as f:
        generic_traceroute_n_q1_ipv4_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/traceroute-n-ipv6-streaming.json'), 'r', encoding='utf-8') as f:
        generic_traceroute_n_ipv6_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/traceroute-long-ipv6-streaming.json'), 'r', encoding='utf-8') as f:
        generic_traceroute_long_ipv6_json = json.loads(f.read())


    def test_traceroute_s_nodata(self):
        """
        Test 'traceroute' with no data
        """
        self.assertEqual(list(jc.parsers.traceroute_s.parse([], quiet=True)), [])

    def test_traceroute_s_noheader(self):
        """
        Test 'traceroute' with missing header row
        """
        self.assertEqual(list(jc.parsers.traceroute_s.parse(self.osx_10_14_6_traceroute_noheader.splitlines(), quiet=True)), self.osx_10_14_6_traceroute_no_header_json)

    def test_traceroute_s_centos_7_7(self):
        """
        Test 'traceroute' on Centos 7.7
        """
        self.assertEqual(list(jc.parsers.traceroute_s.parse(self.centos_7_7_traceroute.splitlines(), quiet=True)), self.centos_7_7_traceroute_json)

    def test_traceroute_s_a_osx_10_14_6(self):
        """
        Test 'traceroute -a' on OSX 10.14.6
        """
        self.assertEqual(list(jc.parsers.traceroute_s.parse(self.osx_10_14_6_traceroute_asn.splitlines(), quiet=True)), self.osx_10_14_6_traceroute_asn_json)

    def test_traceroute_s_mult_addresses_osx_10_14_6(self):
        """
        Test 'traceroute' with multiple addresses returned via dns on OSX 10.14.6
        """
        self.assertEqual(list(jc.parsers.traceroute_s.parse(self.osx_10_14_6_traceroute_mult_addresses.splitlines(), quiet=True)), self.osx_10_14_6_traceroute_mult_addresses_json)

    def test_traceroute_s_q_osx_10_14_6(self):
        """
        Test 'traceroute -q' on OSX 10.14.6
        """
        self.assertEqual(list(jc.parsers.traceroute_s.parse(self.osx_10_14_6_traceroute_q.splitlines(), quiet=True)), self.osx_10_14_6_traceroute_q_json)

    def test_traceroute_s_osx_10_14_6(self):
        """
        Test 'traceroute' on OSX 10.14.6
        """
        self.assertEqual(list(jc.parsers.traceroute_s.parse(self.osx_10_14_6_traceroute.splitlines(), quiet=True)), self.osx_10_14_6_traceroute_json)

    def test_traceroute_s6_mult_addresses_osx_10_14_6(self):
        """
        Test 'traceroute6' with multiple addresses returned via dns on OSX 10.14.6
        """
        self.assertEqual(list(jc.parsers.traceroute_s.parse(self.osx_10_14_6_traceroute6_mult_addresses.splitlines(), quiet=True)), self.osx_10_14_6_traceroute6_mult_addresses_json)

    def test_traceroute_s6_osx_10_14_6(self):
        """
        Test 'traceroute6' on OSX 10.14.6
        """
        self.assertEqual(list(jc.parsers.traceroute_s.parse(self.osx_10_14_6_traceroute6.splitlines(), quiet=True)), self.osx_10_14_6_traceroute6_json)

    def test_traceroute_s_freebsd12(self):
        """
        Test 'traceroute' on freebsd12
        """
        self.assertEqual(list(jc.parsers.traceroute_s.parse(self.freebsd12_traceroute.splitlines(), quiet=True)), self.freebsd12_traceroute_json)

    def test_traceroute_s6_freebsd12(self):
        """
        Test 'traceroute6' on freebsd12
        """
        self.assertEqual(list(jc.parsers.traceroute_s.parse(self.freebsd12_traceroute6.splitlines(), quiet=True)), self.freebsd12_traceroute6_json)

    def test_traceroute_s1_generic(self):
        """
        Test 'traceroute'
        """
        self.assertEqual(list(jc.parsers.traceroute_s.parse(self.generic_traceroute1.splitlines(), quiet=True)), self.generic_traceroute1_json)

    def test_traceroute_s2_generic(self):
        """
        Test 'traceroute'
        """
        self.assertEqual(list(jc.parsers.traceroute_s.parse(self.generic_traceroute2.splitlines(), quiet=True)), self.generic_traceroute2_json)

    def test_traceroute_s3_generic(self):
        """
        Test 'traceroute'
        """
        self.assertEqual(list(jc.parsers.traceroute_s.parse(self.generic_traceroute3.splitlines(), quiet=True)), self.generic_traceroute3_json)

    def test_traceroute_s4_generic(self):
        """
        Test 'traceroute'
        """
        self.assertEqual(list(jc.parsers.traceroute_s.parse(self.generic_traceroute4.splitlines(), quiet=True)), self.generic_traceroute4_json)

    def test_traceroute_s5_generic(self):
        """
        Test 'traceroute'
        """
        self.assertEqual(list(jc.parsers.traceroute_s.parse(self.generic_traceroute5.splitlines(), quiet=True)), self.generic_traceroute5_json)

    def test_traceroute_s6_generic(self):
        """
        Test 'traceroute'
        """
        self.assertEqual(list(jc.parsers.traceroute_s.parse(self.generic_traceroute6.splitlines(), quiet=True)), self.generic_traceroute6_json)

    def test_traceroute_s7_generic(self):
        """
        Test 'traceroute'
        """
        self.assertEqual(list(jc.parsers.traceroute_s.parse(self.generic_traceroute7.splitlines(), quiet=True)), self.generic_traceroute7_json)

    def test_traceroute_s8_generic(self):
        """
        Test 'traceroute'
        """
        self.assertEqual(list(jc.parsers.traceroute_s.parse(self.generic_traceroute8.splitlines(), quiet=True)), self.generic_traceroute8_json)

    def test_traceroute_s_n_ipv4(self):
        """
        Test 'traceroute -n x.x.x.x'
        """
        self.assertEqual(list(jc.parsers.traceroute_s.parse(self.generic_traceroute_n_ipv4.splitlines(), quiet=True)), self.generic_traceroute_n_ipv4_json)

    def test_traceroute_s_n_q1_ipv4(self):
        """
        Test 'traceroute -q1 -n x.x.x.x'
        """
        self.assertEqual(list(jc.parsers.traceroute_s.parse(self.generic_traceroute_n_q1_ipv4.splitlines(), quiet=True)), self.generic_traceroute_n_q1_ipv4_json)

    def test_traceroute_s_n_ipv6(self):
        """
        Test 'traceroute6 -n x::x'
        """
        self.assertEqual(list(jc.parsers.traceroute_s.parse(self.generic_traceroute_n_ipv6.splitlines(), quiet=True)), self.generic_traceroute_n_ipv6_json)

    def test_traceroute_s_long_ipv6(self):
        """
        Test 'traceroute6' with a long ipv6 response
        """
        self.assertEqual(list(jc.parsers.traceroute_s.parse(self.generic_traceroute_long_ipv6.splitlines(), quiet=True)), self.generic_traceroute_long_ipv6_json)


if __name__ == '__main__':
    unittest.main()



import os
import unittest
import json
import jc.parsers.ini

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class MyTests(unittest.TestCase):

    # input
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/ini-test.ini'), 'r', encoding='utf-8') as f:
        generic_ini_test = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/ini-iptelserver.ini'), 'r', encoding='utf-8') as f:
        generic_ini_iptelserver = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/ini-double-quote.ini'), 'r', encoding='utf-8') as f:
        generic_ini_double_quote = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/ini-single-quote.ini'), 'r', encoding='utf-8') as f:
        generic_ini_single_quote = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/ini-mariadb.ini'), 'r', encoding='utf-8') as f:
        generic_ini_mariadb = f.read()

    # output
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/ini-test.json'), 'r', encoding='utf-8') as f:
        generic_ini_test_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/ini-iptelserver.json'), 'r', encoding='utf-8') as f:
        generic_ini_iptelserver_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/ini-double-quote.json'), 'r', encoding='utf-8') as f:
        generic_ini_double_quote_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/ini-single-quote.json'), 'r', encoding='utf-8') as f:
        generic_ini_single_quote_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/ini-mariadb.json'), 'r', encoding='utf-8') as f:
        generic_ini_mariadb_json = json.loads(f.read())


    def test_ini_nodata(self):
        """
        Test the test ini file with no data
        """
        self.assertEqual(jc.parsers.ini.parse('', quiet=True), {})

    def test_ini_test(self):
        """
        Test the test ini file
        """
        self.assertEqual(jc.parsers.ini.parse(self.generic_ini_test, quiet=True), self.generic_ini_test_json)

    def test_ini_iptelserver(self):
        """
        Test the iptelserver ini file
        """
        self.assertEqual(jc.parsers.ini.parse(self.generic_ini_iptelserver, quiet=True), self.generic_ini_iptelserver_json)

    def test_ini_mariadb(self):
        """
        Test the mariadb ini file
        """
        self.assertEqual(jc.parsers.ini.parse(self.generic_ini_mariadb, quiet=True), self.generic_ini_mariadb_json)

    def test_ini_duplicate_keys(self):
        """
        Test input that contains duplicate keys. Only the last value should be used.
        """
        data = '''
[section]
duplicate_key: value1
another_key = foo
duplicate_key = value2
'''
        expected = {'section': {'duplicate_key': 'value2', 'another_key': 'foo'}}
        self.assertEqual(jc.parsers.ini.parse(data, quiet=True), expected)

    def test_ini_missing_top_section(self):
        """
        Test INI file missing top-level section header.
        """
        data = '''
key: value1
another_key = foo
[section2]
key3: bar
key4 =
[section 3]
key5 = "quoted"
'''
        expected = {
            'key': 'value1',
            'another_key': 'foo',
            'section2': {
                'key3': 'bar',
                'key4': ''
            },
            'section 3': {
                'key5': 'quoted'
            }
        }
        self.assertEqual(jc.parsers.ini.parse(data, quiet=True), expected)

    def test_ini_doublequote(self):
        """
        Test ini file with double quotes around a value
        """
        self.assertEqual(jc.parsers.ini.parse(self.generic_ini_double_quote, quiet=True), self.generic_ini_double_quote_json)

    def test_ini_singlequote(self):
        """
        Test ini file with single quotes around a value
        """
        self.assertEqual(jc.parsers.ini.parse(self.generic_ini_single_quote, quiet=True), self.generic_ini_single_quote_json)

    def test_ini_single_key_no_value(self):
        """
        Test ini file with a single item with no value. This caused issues in jc v.1.25.0
        """
        data = '''[data]
novalue
'''
        expected = {"data":{"novalue":""}}
        self.assertEqual(jc.parsers.ini.parse(data, quiet=True), expected)


if __name__ == '__main__':
    unittest.main()



import os
import json
import unittest
import jc.parsers.efibootmgr

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class MyTests(unittest.TestCase):

    # input
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/opensuse-leap-15.5/efibootmgr-v.out'), 'r', encoding='utf-8') as f:
        opensuse_leap_15_5_efibootmgr_v = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/opensuse-leap-15.5/efibootmgr.out'), 'r', encoding='utf-8') as f:
        opensuse_leap_15_5_efibootmgr = f.read()

    # output
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/opensuse-leap-15.5/efibootmgr-v.json'), 'r', encoding='utf-8') as f:
        opensuse_leap_15_5_efibootmgr_json_v = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/opensuse-leap-15.5/efibootmgr.json'), 'r', encoding='utf-8') as f:
        opensuse_leap_15_5_efibootmgr_json = json.loads(f.read())

    def test_efibootmgr_nodata(self):
        """
        Test 'efibootmgr' with no data
        """
        self.assertEqual(jc.parsers.efibootmgr.parse('', quiet=True), {})

    def test_efibootmgr_v_opensuse_leap_15_5(self):
        """
        Test 'efibootmgr -v' on Opensuse Leap 15.5
        """
        self.assertEqual(jc.parsers.efibootmgr.parse(self.opensuse_leap_15_5_efibootmgr_v, quiet=True), self.opensuse_leap_15_5_efibootmgr_json_v)

    def test_efibootmgr_opensuse_leap_15_5(self):
        """
        Test 'efibootmgr' on Opensuse Leap 15.5
        """
        self.assertEqual(jc.parsers.efibootmgr.parse(self.opensuse_leap_15_5_efibootmgr, quiet=True), self.opensuse_leap_15_5_efibootmgr_json)


if __name__ == '__main__':
    unittest.main()



import os
import json
import unittest
import jc.parsers.cef_s

THIS_DIR = os.path.dirname(os.path.abspath(__file__))

# To create streaming output use:
# $ cat cef.out | jc --cef-s | jello -c > cef-streaming.json


class MyTests(unittest.TestCase):

    # input
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/cef.out'), 'r', encoding='utf-8') as f:
        cef = f.read()

    # output
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/cef-streaming.json'), 'r', encoding='utf-8') as f:
        cef_streaming_json = json.loads(f.read())

    def test_cef_s_nodata(self):
        """
        Test 'cef' with no data
        """
        self.assertEqual(list(jc.parsers.cef_s.parse([], quiet=True)), [])

    def test_cef_s_sample(self):
        """
        Test with sample cef log
        """
        self.assertEqual(list(jc.parsers.cef_s.parse(self.cef.splitlines(), quiet=True)), self.cef_streaming_json)


if __name__ == '__main__':
    unittest.main()



import os
import json
import unittest
import jc.parsers.netstat

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class MyTests(unittest.TestCase):

    #
    # input
    #

    # netstat
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/netstat.out'), 'r', encoding='utf-8') as f:
        centos_7_7_netstat = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-18.04/netstat.out'), 'r', encoding='utf-8') as f:
        ubuntu_18_4_netstat = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/netstat-l.out'), 'r', encoding='utf-8') as f:
        centos_7_7_netstat_l = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-18.04/netstat-l.out'), 'r', encoding='utf-8') as f:
        ubuntu_18_4_netstat_l = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/netstat-p.out'), 'r', encoding='utf-8') as f:
        centos_7_7_netstat_p = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-18.04/netstat-p.out'), 'r', encoding='utf-8') as f:
        ubuntu_18_4_netstat_p = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/netstat-sudo-lnp.out'), 'r', encoding='utf-8') as f:
        centos_7_7_netstat_sudo_lnp = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-18.04/netstat-sudo-lnp.out'), 'r', encoding='utf-8') as f:
        ubuntu_18_4_netstat_sudo_lnp = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-18.04/netstat-sudo-lnp-space.out'), 'r', encoding='utf-8') as f:
        ubuntu_18_4_netstat_sudo_lnp_space = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/netstat-sudo-aeep.out'), 'r', encoding='utf-8') as f:
        centos_7_7_netstat_sudo_aeep = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-18.04/netstat-sudo-aeep.out'), 'r', encoding='utf-8') as f:
        ubuntu_18_4_netstat_sudo_aeep = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/fedora32/netstat.out'), 'r', encoding='utf-8') as f:
        fedora32_netstat = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/netstat.out'), 'r', encoding='utf-8') as f:
        osx_14_6_netstat = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/netstat-An.out'), 'r', encoding='utf-8') as f:
        osx_14_6_netstat_An = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/netstat-Abn.out'), 'r', encoding='utf-8') as f:
        osx_14_6_netstat_Abn = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/freebsd12/netstat-Aa.out'), 'r', encoding='utf-8') as f:
        freebsd12_netstat_Aa = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/freebsd12/netstat-an.out'), 'r', encoding='utf-8') as f:
        freebsd12_netstat_an = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/freebsd12/netstat-AanP.out'), 'r', encoding='utf-8') as f:
        freebsd12_netstat_AanP = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/freebsd12/netstat-AaT.out'), 'r', encoding='utf-8') as f:
        freebsd12_netstat_AaT = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/freebsd12/netstat-Aax.out'), 'r', encoding='utf-8') as f:
        freebsd12_netstat_Aax = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/freebsd12/netstat-aT.out'), 'r', encoding='utf-8') as f:
        freebsd12_netstat_aT = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/netstat-old.out'), 'r', encoding='utf-8') as f:
        generic_netstat_old = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/netstat-no-state.out'), 'r', encoding='utf-8') as f:
        generic_netstat_no_state = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/windows/windows-10/netstat.out'), 'r', encoding='utf-8') as f:
        windows_netstat = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/windows/windows-10/netstat-an.out'), 'r', encoding='utf-8') as f:
        windows_netstat_an = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/windows/windows-10/netstat-aon.out'), 'r', encoding='utf-8') as f:
        windows_netstat_aon = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/windows/windows-10/netstat-aonb.out'), 'r', encoding='utf-8') as f:
        windows_netstat_aonb = f.read()

    # netstat -r
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/netstat-r.out'), 'r', encoding='utf-8') as f:
        centos_7_7_netstat_r = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/netstat-rne.out'), 'r', encoding='utf-8') as f:
        centos_7_7_netstat_rne = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/netstat-rnee.out'), 'r', encoding='utf-8') as f:
        centos_7_7_netstat_rnee = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-18.04/netstat-r.out'), 'r', encoding='utf-8') as f:
        ubuntu_18_4_netstat_r = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-18.04/netstat-rne.out'), 'r', encoding='utf-8') as f:
        ubuntu_18_4_netstat_rne = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-18.04/netstat-rnee.out'), 'r', encoding='utf-8') as f:
        ubuntu_18_4_netstat_rnee = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/netstat-r.out'), 'r', encoding='utf-8') as f:
        osx_14_6_netstat_r = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/netstat-rnl.out'), 'r', encoding='utf-8') as f:
        osx_14_6_netstat_rnl = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/freebsd12/netstat-r.out'), 'r', encoding='utf-8') as f:
        freebsd12_netstat_r = f.read()

    # netstat -i
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/netstat-i.out'), 'r', encoding='utf-8') as f:
        centos_7_7_netstat_i = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-18.04/netstat-i.out'), 'r', encoding='utf-8') as f:
        ubuntu_18_4_netstat_i = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/netstat-i.out'), 'r', encoding='utf-8') as f:
        osx_14_6_netstat_i = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/freebsd12/netstat-i.out'), 'r', encoding='utf-8') as f:
        freebsd12_netstat_i = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/freebsd12/netstat-ib.out'), 'r', encoding='utf-8') as f:
        freebsd12_netstat_ib = f.read()

    #
    # output
    #

    # netstat
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/netstat.json'), 'r', encoding='utf-8') as f:
        centos_7_7_netstat_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-18.04/netstat.json'), 'r', encoding='utf-8') as f:
        ubuntu_18_4_netstat_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/netstat-l.json'), 'r', encoding='utf-8') as f:
        centos_7_7_netstat_l_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-18.04/netstat-l.json'), 'r', encoding='utf-8') as f:
        ubuntu_18_4_netstat_l_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/netstat-p.json'), 'r', encoding='utf-8') as f:
        centos_7_7_netstat_p_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-18.04/netstat-p.json'), 'r', encoding='utf-8') as f:
        ubuntu_18_4_netstat_p_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/netstat-sudo-lnp.json'), 'r', encoding='utf-8') as f:
        centos_7_7_netstat_sudo_lnp_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-18.04/netstat-sudo-lnp.json'), 'r', encoding='utf-8') as f:
        ubuntu_18_4_netstat_sudo_lnp_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-18.04/netstat-sudo-lnp-space.json'), 'r', encoding='utf-8') as f:
        ubuntu_18_4_netstat_sudo_lnp_space_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/netstat-sudo-aeep.json'), 'r', encoding='utf-8') as f:
        centos_7_7_netstat_sudo_aeep_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-18.04/netstat-sudo-aeep.json'), 'r', encoding='utf-8') as f:
        ubuntu_18_4_netstat_sudo_aeep_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/fedora32/netstat.json'), 'r', encoding='utf-8') as f:
        fedora32_netstat_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/netstat.json'), 'r', encoding='utf-8') as f:
        osx_14_6_netstat_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/netstat-An.json'), 'r', encoding='utf-8') as f:
        osx_14_6_netstat_An_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/netstat-Abn.json'), 'r', encoding='utf-8') as f:
        osx_14_6_netstat_Abn_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/freebsd12/netstat-Aa.json'), 'r', encoding='utf-8') as f:
        freebsd12_netstat_Aa_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/freebsd12/netstat-AanP.json'), 'r', encoding='utf-8') as f:
        freebsd12_netstat_AanP_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/freebsd12/netstat-AaT.json'), 'r', encoding='utf-8') as f:
        freebsd12_netstat_AaT_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/freebsd12/netstat-Aax.json'), 'r', encoding='utf-8') as f:
        freebsd12_netstat_Aax_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/freebsd12/netstat-aT.json'), 'r', encoding='utf-8') as f:
        freebsd12_netstat_aT_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/freebsd12/netstat-an.json'), 'r', encoding='utf-8') as f:
        freebsd12_netstat_an_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/netstat-old.json'), 'r', encoding='utf-8') as f:
        generic_netstat_old_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/netstat-no-state.json'), 'r', encoding='utf-8') as f:
        generic_netstat_no_state_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/windows/windows-10/netstat.json'), 'r', encoding='utf-8') as f:
        windows_netstat_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/windows/windows-10/netstat-an.json'), 'r', encoding='utf-8') as f:
        windows_netstat_an_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/windows/windows-10/netstat-aon.json'), 'r', encoding='utf-8') as f:
        windows_netstat_aon_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/windows/windows-10/netstat-aonb.json'), 'r', encoding='utf-8') as f:
        windows_netstat_aonb_json = json.loads(f.read())

    # netsat -r
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/netstat-r.json'), 'r', encoding='utf-8') as f:
        centos_7_7_netstat_r_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/netstat-rne.json'), 'r', encoding='utf-8') as f:
        centos_7_7_netstat_rne_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/netstat-rnee.json'), 'r', encoding='utf-8') as f:
        centos_7_7_netstat_rnee_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-18.04/netstat-r.json'), 'r', encoding='utf-8') as f:
        ubuntu_18_4_netstat_r_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-18.04/netstat-rne.json'), 'r', encoding='utf-8') as f:
        ubuntu_18_4_netstat_rne_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-18.04/netstat-rnee.json'), 'r', encoding='utf-8') as f:
        ubuntu_18_4_netstat_rnee_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/netstat-r.json'), 'r', encoding='utf-8') as f:
        osx_14_6_netstat_r_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/netstat-rnl.json'), 'r', encoding='utf-8') as f:
        osx_14_6_netstat_rnl_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/freebsd12/netstat-r.json'), 'r', encoding='utf-8') as f:
        freebsd12_netstat_r_json = json.loads(f.read())

    # netstat -i
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/netstat-i.json'), 'r', encoding='utf-8') as f:
        centos_7_7_netstat_i_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-18.04/netstat-i.json'), 'r', encoding='utf-8') as f:
        ubuntu_18_4_netstat_i_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/netstat-i.json'), 'r', encoding='utf-8') as f:
        osx_14_6_netstat_i_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/freebsd12/netstat-i.json'), 'r', encoding='utf-8') as f:
        freebsd12_netstat_i_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/freebsd12/netstat-ib.json'), 'r', encoding='utf-8') as f:
        freebsd12_netstat_ib_json = json.loads(f.read())


    def test_netstat_nodata(self):
        """
        Test 'netstat' with no data
        """
        self.assertEqual(jc.parsers.netstat.parse('', quiet=True), [])

    def test_netstat_centos_7_7(self):
        """
        Test 'netstat' on Centos 7.7
        """
        self.assertEqual(jc.parsers.netstat.parse(self.centos_7_7_netstat, quiet=True), self.centos_7_7_netstat_json)

    def test_netstat_ubuntu_18_4(self):
        """
        Test 'netstat' on Ubuntu 18.4
        """
        self.assertEqual(jc.parsers.netstat.parse(self.ubuntu_18_4_netstat, quiet=True), self.ubuntu_18_4_netstat_json)

    def test_netstat_l_centos_7_7(self):
        """
        Test 'netstat -l' on Centos 7.7
        """
        self.assertEqual(jc.parsers.netstat.parse(self.centos_7_7_netstat_l, quiet=True), self.centos_7_7_netstat_l_json)

    def test_netstat_l_ubuntu_18_4(self):
        """
        Test 'netstat -l' on Ubuntu 18.4
        """
        self.assertEqual(jc.parsers.netstat.parse(self.ubuntu_18_4_netstat_l, quiet=True), self.ubuntu_18_4_netstat_l_json)

    def test_netstat_p_centos_7_7(self):
        """
        Test 'netstat -l' on Centos 7.7
        """
        self.assertEqual(jc.parsers.netstat.parse(self.centos_7_7_netstat_p, quiet=True), self.centos_7_7_netstat_p_json)

    def test_netstat_p_ubuntu_18_4(self):
        """
        Test 'netstat -l' on Ubuntu 18.4
        """
        self.assertEqual(jc.parsers.netstat.parse(self.ubuntu_18_4_netstat_p, quiet=True), self.ubuntu_18_4_netstat_p_json)

    def test_netstat_sudo_lnp_centos_7_7(self):
        """
        Test 'sudo netstat -lnp' on Centos 7.7
        """
        self.assertEqual(jc.parsers.netstat.parse(self.centos_7_7_netstat_sudo_lnp, quiet=True), self.centos_7_7_netstat_sudo_lnp_json)

    def test_netstat_sudo_lnp_ubuntu_18_4(self):
        """
        Test 'sudo netstat -lnp' on Ubuntu 18.4
        """
        self.assertEqual(jc.parsers.netstat.parse(self.ubuntu_18_4_netstat_sudo_lnp, quiet=True), self.ubuntu_18_4_netstat_sudo_lnp_json)

    def test_netstat_sudo_lnp_ubuntu_18_4(self):
        """
        Test 'sudo netstat -lnp' on Ubuntu 18.4 with a space in the process name(special case)
        """
        self.assertEqual(jc.parsers.netstat.parse(self.ubuntu_18_4_netstat_sudo_lnp_space, quiet=True), self.ubuntu_18_4_netstat_sudo_lnp_space_json)

    def test_netstat_sudo_aeep_centos_7_7(self):
        """
        Test 'sudo netstat -aeep' on Centos 7.7
        """
        self.assertEqual(jc.parsers.netstat.parse(self.centos_7_7_netstat_sudo_aeep, quiet=True), self.centos_7_7_netstat_sudo_aeep_json)

    def test_netstat_sudo_aeep_ubuntu_18_4(self):
        """
        Test 'sudo netstat -aeep' on Ubuntu 18.4
        """
        self.assertEqual(jc.parsers.netstat.parse(self.ubuntu_18_4_netstat_sudo_aeep, quiet=True), self.ubuntu_18_4_netstat_sudo_aeep_json)

    def test_netstat_fedora32(self):
        """
        Test 'netstat' on Fedora32
        """
        self.assertEqual(jc.parsers.netstat.parse(self.fedora32_netstat, quiet=True), self.fedora32_netstat_json)

    def test_netstat_osx_14_6(self):
        """
        Test 'netstat' on OSX 14.6
        """
        self.assertEqual(jc.parsers.netstat.parse(self.osx_14_6_netstat, quiet=True), self.osx_14_6_netstat_json)

    def test_netstat_An_osx_14_6(self):
        """
        Test 'netstat -An' on OSX 14.6
        """
        self.assertEqual(jc.parsers.netstat.parse(self.osx_14_6_netstat_An, quiet=True), self.osx_14_6_netstat_An_json)

    def test_netstat_Abn_osx_14_6(self):
        """
        Test 'netstat -Abn' on OSX 14.6
        """
        self.assertEqual(jc.parsers.netstat.parse(self.osx_14_6_netstat_Abn, quiet=True), self.osx_14_6_netstat_Abn_json)

    def test_netstat_Aa_freebsd12(self):
        """
        Test 'netstat -Aa' on FreeBSD12
        """
        self.assertEqual(jc.parsers.netstat.parse(self.freebsd12_netstat_Aa, quiet=True), self.freebsd12_netstat_Aa_json)

    def test_netstat_AanP_freebsd12(self):
        """
        Test 'netstat -AanP' on FreeBSD12
        """
        self.assertEqual(jc.parsers.netstat.parse(self.freebsd12_netstat_AanP, quiet=True), self.freebsd12_netstat_AanP_json)

    def test_netstat_AaT_freebsd12(self):
        """
        Test 'netstat -AaT' on FreeBSD12
        """
        self.assertEqual(jc.parsers.netstat.parse(self.freebsd12_netstat_AaT, quiet=True), self.freebsd12_netstat_AaT_json)

    def test_netstat_Aax_freebsd12(self):
        """
        Test 'netstat -Aax' on FreeBSD12
        """
        self.assertEqual(jc.parsers.netstat.parse(self.freebsd12_netstat_Aax, quiet=True), self.freebsd12_netstat_Aax_json)

    def test_netstat_aT_freebsd12(self):
        """
        Test 'netstat -aT' on FreeBSD12
        """
        self.assertEqual(jc.parsers.netstat.parse(self.freebsd12_netstat_aT, quiet=True), self.freebsd12_netstat_aT_json)

    def test_netstat_an_freebsd12(self):
        """
        Test 'netstat -an' on FreeBSD12
        """
        self.assertEqual(jc.parsers.netstat.parse(self.freebsd12_netstat_an, quiet=True), self.freebsd12_netstat_an_json)

    def test_netstat_old_generic(self):
        """
        Test 'netstat' with older version of netstat on linux
        """
        self.assertEqual(jc.parsers.netstat.parse(self.generic_netstat_old, quiet=True), self.generic_netstat_old_json)

    def test_netstat_no_state_generic(self):
        """
        Test 'netstat' with no state in network output
        """
        self.assertEqual(jc.parsers.netstat.parse(self.generic_netstat_no_state, quiet=True), self.generic_netstat_no_state_json)

    def test_netstat_r_centos_7_7(self):
        """
        Test 'netstat -r' on Centos 7.7
        """
        self.assertEqual(jc.parsers.netstat.parse(self.centos_7_7_netstat_r, quiet=True), self.centos_7_7_netstat_r_json)

    def test_netstat_rne_centos_7_7(self):
        """
        Test 'netstat -rne' on Centos 7.7
        """
        self.assertEqual(jc.parsers.netstat.parse(self.centos_7_7_netstat_rne, quiet=True), self.centos_7_7_netstat_rne_json)

    def test_netstat_rnee_centos_7_7(self):
        """
        Test 'netstat -rnee' on Centos 7.7
        """
        self.assertEqual(jc.parsers.netstat.parse(self.centos_7_7_netstat_rnee, quiet=True), self.centos_7_7_netstat_rnee_json)

    def test_netstat_r_ubuntu_18_4(self):
        """
        Test 'netstat -r' on Ubuntu 18.4
        """
        self.assertEqual(jc.parsers.netstat.parse(self.ubuntu_18_4_netstat_r, quiet=True), self.ubuntu_18_4_netstat_r_json)

    def test_netstat_rne_ubuntu_18_4(self):
        """
        Test 'netstat -rne' on Ubuntu 18.4
        """
        self.assertEqual(jc.parsers.netstat.parse(self.ubuntu_18_4_netstat_rne, quiet=True), self.ubuntu_18_4_netstat_rne_json)

    def test_netstat_rnee_ubuntu_18_4(self):
        """
        Test 'netstat -rnee' on Ubuntu 18.4
        """
        self.assertEqual(jc.parsers.netstat.parse(self.ubuntu_18_4_netstat_rnee, quiet=True), self.ubuntu_18_4_netstat_rnee_json)

    def test_netstat_r_osx_14_6(self):
        """
        Test 'netstat -r' on OSX 14.6
        """
        self.assertEqual(jc.parsers.netstat.parse(self.osx_14_6_netstat_r, quiet=True), self.osx_14_6_netstat_r_json)

    def test_netstat_rnl_osx_14_6(self):
        """
        Test 'netstat -rnl' on OSX 14.6
        """
        self.assertEqual(jc.parsers.netstat.parse(self.osx_14_6_netstat_rnl, quiet=True), self.osx_14_6_netstat_rnl_json)

    def test_netstat_r_freebsd12(self):
        """
        Test 'netstat -r' on FreeBSD12
        """
        self.assertEqual(jc.parsers.netstat.parse(self.freebsd12_netstat_r, quiet=True), self.freebsd12_netstat_r_json)

    def test_netstat_i_centos_7_7(self):
        """
        Test 'netstat -i' on Centos 7.7
        """
        self.assertEqual(jc.parsers.netstat.parse(self.centos_7_7_netstat_i, quiet=True), self.centos_7_7_netstat_i_json)

    def test_netstat_i_ubuntu_18_4(self):
        """
        Test 'netstat -i' on Ubuntu 18.4
        """
        self.assertEqual(jc.parsers.netstat.parse(self.ubuntu_18_4_netstat_i, quiet=True), self.ubuntu_18_4_netstat_i_json)

    def test_netstat_i_osx_14_6(self):
        """
        Test 'netstat -i' on OSX 14.6
        """
        self.assertEqual(jc.parsers.netstat.parse(self.osx_14_6_netstat_i, quiet=True), self.osx_14_6_netstat_i_json)

    def test_netstat_i_freebsd12(self):
        """
        Test 'netstat -i' on FreeBSD12
        """
        self.assertEqual(jc.parsers.netstat.parse(self.freebsd12_netstat_i, quiet=True), self.freebsd12_netstat_i_json)

    def test_netstat_ib_freebsd12(self):
        """
        Test 'netstat -ib' on FreeBSD12
        """
        self.assertEqual(jc.parsers.netstat.parse(self.freebsd12_netstat_ib, quiet=True), self.freebsd12_netstat_ib_json)

    def test_netstat_windows(self):
        """
        Test 'netstat' on Windows
        """
        self.assertEqual(jc.parsers.netstat.parse(self.windows_netstat, quiet=True), self.windows_netstat_json)

    def test_netstat_an_windows(self):
        """
        Test 'netstat -an' on Windows
        """
        self.assertEqual(jc.parsers.netstat.parse(self.windows_netstat_an, quiet=True), self.windows_netstat_an_json)

    def test_netstat_aon_windows(self):
        """
        Test 'netstat -aon' on Windows
        """
        self.assertEqual(jc.parsers.netstat.parse(self.windows_netstat_aon, quiet=True), self.windows_netstat_aon_json)

    def test_netstat_aonb_windows(self):
        """
        Test 'netstat -aonb' on Windows
        """
        self.assertEqual(jc.parsers.netstat.parse(self.windows_netstat_aonb, quiet=True), self.windows_netstat_aonb_json)

if __name__ == '__main__':
    unittest.main()



r"""jc - JSON Convert `/proc/<pid>/numa_maps` file parser

This parser will attempt to convert number values to integers. If that is
not desired, please use the `--raw` option (cli) or `raw=True` argument
(module).

Usage (cli):

    $ cat /proc/1/numa_maps | jc --proc

or

    $ jc /proc/1/numa_maps

or

    $ cat /proc/1/numa_maps | jc --proc-pid-numa-maps

Usage (module):

    import jc
    result = jc.parse('proc', proc_pid_numa_maps_file)

or

    import jc
    result = jc.parse('proc_pid_numa_maps', proc_pid_numa_maps_file)

Schema:

Integer conversion for Key/value pairs will be attempted.

    [
      {
        "address":                    string,
        "policy":                     string,
        "<key>":                      string/integer,
        "options": [
                                      string  # [0]
        ]
      }
    ]

    [0] remaining individual words that are not part of a key/value pair

Examples:

    $ cat /proc/1/numa_maps | jc --proc -p
    [
      {
        "address": "7f53b5083000",
        "policy": "default",
        "file": "/usr/lib/x86_64-linux-gnu/ld-2.32.so",
        "anon": 2,
        "dirty": 2,
        "N0": 2,
        "kernelpagesize_kB": 4
      },
      {
        "address": "7ffd1b23e000",
        "policy": "default",
        "anon": 258,
        "dirty": 258,
        "N0": 258,
        "kernelpagesize_kB": 4,
        "options": [
          "stack"
        ]
      },
      ...
    ]

    $ cat /proc/1/numa_maps | jc --proc-pid-numa-maps -p -r
    [
      {
        "address": "7f53b5083000",
        "policy": "default",
        "file": "/usr/lib/x86_64-linux-gnu/ld-2.32.so",
        "anon": "2",
        "dirty": "2",
        "N0": "2",
        "kernelpagesize_kB": "4"
      },
      {
        "address": "7ffd1b23e000",
        "policy": "default",
        "anon": "258",
        "dirty": "258",
        "N0": "258",
        "kernelpagesize_kB": "4",
        "options": [
          "stack"
        ]
      },
      ...
    ]
"""
from typing import List, Dict
import jc.utils
from jc.parsers.universal import simple_table_parse


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.0'
    description = '`/proc/<pid>/numa_maps` file parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux']
    tags = ['file']
    hidden = True


__version__ = info.version


def _process(proc_data: List[Dict]) -> List[Dict]:
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        List of Dictionaries. Structured to conform to the schema.
    """
    for entry in proc_data:
        for key, val in entry.items():
            try:
                entry[key] = int(val)
            except Exception:
                pass

    return proc_data


def parse(
    data: str,
    raw: bool = False,
    quiet: bool = False
) -> List[Dict]:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output: List = []

    if jc.utils.has_data(data):

        header = 'address policy details\n'
        data = header + data

        raw_output = simple_table_parse(data.splitlines())

        for row in raw_output:
            if 'details' in row:
                detail_split = row['details'].split()

                options = []
                for item in detail_split:
                    if '=' in item:
                        key, val = item.split('=')
                        row.update({key: val})
                    else:
                        options.append(item)

                if options:
                    row['options'] = options

                del row['details']

    return raw_output if raw else _process(raw_output)



r"""jc - JSON Convert `xrandr` command output parser

Usage (cli):

    $ xrandr | jc --xrandr
    $ xrandr --properties | jc --xrandr

or

    $ jc xrandr

Usage (module):

    import jc
    result = jc.parse('xrandr', xrandr_command_output)

Schema:

    {
      "screens": [
        {
          "screen_number":                     integer,
          "minimum_width":                     integer,
          "minimum_height":                    integer,
          "current_width":                     integer,
          "current_height":                    integer,
          "maximum_width":                     integer,
          "maximum_height":                    integer,
          "devices": [
            {
              "props": {

                # either an EdidModel object or key value pairs
                <key_name> : {
                  "name": string,
                  "product_id": string,
                  "serial_number": string,
                }

              },
              "resolution_modes": [
                {
                  "resolution_width":          integer,
                  "resolution_height":         integer,
                  "is_high_resolution":        boolean,
                  "frequencies": [
                    {
                      "frequency":             float,
                      "is_current":            boolean,
                      "is_preferred":          boolean
                    }
                  ]
                }
              ]
            }
          ],
          "is_connected":                      boolean,
          "is_primary":                        boolean,
          "device_name":                       string,
          "model_name":                        string,
          "product_id"                         string,
          "serial_number":                     string,
          "resolution_width":                  integer,
          "resolution_height":                 integer,
          "offset_width":                      integer,
          "offset_height":                     integer,
          "dimension_width":                   integer,
          "dimension_height":                  integer,
          "rotation":                          string,
          "reflection":                        string
        }
      ]
    }

Examples:

    $ xrandr | jc --xrandr -p
    {
      "screens": [
        {
          "screen_number": 0,
          "minimum_width": 8,
          "minimum_height": 8,
          "current_width": 1920,
          "current_height": 1080,
          "maximum_width": 32767,
          "maximum_height": 32767,
          "devices": [
            {
              "resolution_modes": [
                {
                  "resolution_width": 1920,
                  "resolution_height": 1080,
                  "is_high_resolution": false,
                  "frequencies": [
                    {
                      "frequency": 60.03,
                      "is_current": true,
                      "is_preferred": true
                    }
                  ]
                },
              ],
              "is_connected": true,
              "is_primary": true,
              "device_name": "eDP1",
              "resolution_width": 1920,
              "resolution_height": 1080,
              "offset_width": 0,
              "offset_height": 0,
              "dimension_width": 310,
              "dimension_height": 170,
              "rotation": "normal",
              "reflection": "normal"
            }
          ]
        }
      ]
    }

    $ xrandr --properties | jc --xrandr -p
    {
      "screens": [
        {
          "devices": [
            {
              "props": {
                "EDID": [
                  "00ffff5700000000",
                  "001c01a5544d9a27",
                  "0e50540101010101",
                  "010101383e401010",
                  "3500350f00000000",
                  "0000000000fe0041",
                  "554f0a20000000fe",
                  "00423137200a0070"
                ],
                "scaling mode": [
                  "Full aspect ",
                  "supported: Full, Center, Full aspect"
                ],
                "Colorspace": [
                  "Default ",
                  "supported: Default, RGB_Wide_Gamut_Fixed_Point, RGB_Wide_Gamut_Floating_Point, opRGB, DCI-P3_RGB_D65, BT2020_RGB, BT601_YCC, BT709_YCC, XVYCC_601, XVYCC_709, SYCC_601, opYCC_601, BT2020_CYCC, BT2020_YCC"
                ],
                "EdidModel": {
                  "name": "Generic",
                  "product_id": "22333",
                  "serial_number": "0"
                }
              },
              "resolution_modes": [
                {
                  "resolution_width": 320,
                  "resolution_height": 180,
                  "is_high_resolution": false,
                  "frequencies": [
                    {
                      "frequency": 59.84,
                      "is_current": false,
                      "is_preferred": false
                    }
                  ]
                }
              ],
              "is_connected": true,
              "is_primary": true,
              "device_name": "eDP-1",
              "rotation": "normal",
              "reflection": "normal",
              "resolution_width": 1920,
              "resolution_height": 1080,
              "offset_width": 0,
              "offset_height": 0,
              "dimension_width": 309,
              "dimension_height": 174
            }
          ],
          "screen_number": 0,
          "minimum_width": 320,
          "minimum_height": 200,
          "current_width": 1920,
          "current_height": 1080,
          "maximum_width": 16384,
          "maximum_height": 16384
        }
      ]
    }
"""
from collections import defaultdict
from enum import Enum
import re
from typing import Dict, List, Tuple, Union

import jc.utils
from jc.parsers.pyedid.edid import Edid
from jc.parsers.pyedid.helpers.edid_helper import EdidHelper

Match = None
try:
    # Added Python 3.7
    Match = re.Match
except AttributeError:
    Match = type(re.match("", ""))


class info:
    """Provides parser metadata (version, author, etc.)"""

    version = "2.1"
    description = "`xrandr` command parser"
    author = "Kevin Lyter"
    author_email = "code (at) lyterk.com"
    details = "Using parts of the pyedid library at https://github.com/jojonas/pyedid."
    compatible = ["linux", "darwin", "cygwin", "aix", "freebsd"]
    magic_commands = ["xrandr"]
    tags = ["command"]


__version__ = info.version

# NOTE: When developing, comment out the try statement and catch block to get
# TypedDict type hints and valid type errors.
try:
    # Added in Python 3.8
    from typing import TypedDict

    Frequency = TypedDict(
        "Frequency",
        {
            "frequency": float,
            "is_current": bool,
            "is_preferred": bool,
        },
    )
    ResolutionMode = TypedDict(
        "ResolutionMode",
        {
            "resolution_width": int,
            "resolution_height": int,
            "is_high_resolution": bool,
            "frequencies": List[Frequency],
        },
    )
    EdidModel = TypedDict(
        "EdidModel",
        {
            "name": str,
            "product_id": str,
            "serial_number": str,
        },
    )
    Props = Dict[str, Union[List[str], EdidModel]]
    Device = TypedDict(
        "Device",
        {
            "device_name": str,
            "model_name": str,
            "product_id": str,
            "serial_number": str,
            "is_connected": bool,
            "is_primary": bool,
            "resolution_width": int,
            "resolution_height": int,
            "offset_width": int,
            "offset_height": int,
            "dimension_width": int,
            "dimension_height": int,
            "props": Props,
            "resolution_modes": List[ResolutionMode],
            "rotation": str,
            "reflection": str,
        },
    )
    Screen = TypedDict(
        "Screen",
        {
            "screen_number": int,
            "minimum_width": int,
            "minimum_height": int,
            "current_width": int,
            "current_height": int,
            "maximum_width": int,
            "maximum_height": int,
            "devices": List[Device],
        },
    )
    Response = TypedDict(
        "Response",
        {
            "screens": List[Screen],
        },
    )
except ImportError:
    EdidModel = Dict[str, str]
    Props = Dict[str, Union[List[str], EdidModel]]
    Frequency = Dict[str, Union[float, bool]]
    ResolutionMode = Dict[str, Union[int, bool, List[Frequency]]]
    Device = Dict[str, Union[str, int, bool, List[ResolutionMode]]]
    Screen = Dict[str, Union[int, List[Device]]]
    Response = Dict[str, Screen]


_screen_pattern = (
    r"Screen (?P<screen_number>\d+): "
    + r"minimum (?P<minimum_width>\d+) x (?P<minimum_height>\d+), "
    + r"current (?P<current_width>\d+) x (?P<current_height>\d+), "
    + r"maximum (?P<maximum_width>\d+) x (?P<maximum_height>\d+)"
)


# eDP1 connected primary 1920x1080+0+0 (normal left inverted right x axis y axis)
#       310mm x 170mm
# regex101 demo link
_device_pattern = (
    r"(?P<device_name>.+) "
    + r"(?P<is_connected>(connected|disconnected)) ?"
    + r"(?P<is_primary> primary)? ?"
    + r"((?P<resolution_width>\d+)x(?P<resolution_height>\d+)"
    + r"\+(?P<offset_width>\d+)\+(?P<offset_height>\d+))? "
    + r"(?P<rotation>(normal|right|left|inverted)?) ?"
    + r"(?P<reflection>(X axis|Y axis|X and Y axis)?) ?"
    + r"(\(normal left inverted right x axis y axis\))?"
    + r"( ?((?P<dimension_width>\d+)mm x (?P<dimension_height>\d+)mm)?)?"
)

# 1920x1080i     60.03*+  59.93
# 1920x1080     60.00 +  50.00    59.94
_resolution_mode_pattern = r"\s*(?P<resolution_width>\d+)x(?P<resolution_height>\d+)(?P<is_high_resolution>i)?\s+(?P<rest>.*)"
_frequencies_pattern = r"(((?P<frequency>\d+\.\d+)(?P<star>\*| |)(?P<plus>\+?)?)+)"


# Values sometimes appear on the same lines as the keys (CscMatrix), sometimes on the line
# below (as with EDIDs), and sometimes both (CTM).
# Capture the key line that way.
#
# CTM: 0 1 0 0 0 0 0 0 0 1 0 0 0 0 0 0
#         0 1
# CscMatrix: 65536 0 0 0 0 65536 0 0 0 0 65536 0
# EDID:
#         00ffffffffffff0010ac33424c303541
#         0f210104b53c22783eee95a3544c9926
_prop_key_pattern = r"\s+(?P<key>[\w| |\-|_]+):\s?(?P<maybe_value>.*)"


class LineType(Enum):
    Screen = 1
    Device = 2
    ResolutionMode = 3
    PropKey = 4
    PropValue = 5
    Invalid = 6


class _Line:
    """Provide metadata about line to make handling it more simple across fn boundaries"""

    def __init__(self, s: str, t: LineType, m: Match):
        self.s = s
        self.t = t
        self.m = m

    @classmethod
    def categorize(cls, line: str) -> "_Line":
        """Iterate through line char by char to see what type of line it is. Apply regexes for more distinctness. Save the regexes and return them for later processing."""
        i = 0
        tab_count = 0
        while True:
            try:
                c = line[i]
            except:
                # Really shouldn't be getting to the end of the line
                raise Exception(f"Reached end of line unexpectedly: '{line}'")

            if not c.isspace():
                if tab_count == 0:
                    screen_match = re.match(_screen_pattern, line)
                    if screen_match:
                        return cls(line, LineType.Screen, screen_match)

                    device_match = re.match(_device_pattern, line)
                    if device_match:
                        return cls(line, LineType.Device, device_match)
                    else:
                        break
                elif tab_count == 1:
                    match = re.match(_prop_key_pattern, line)
                    if match:
                        return cls(line, LineType.PropKey, match)
                    else:
                        break
                else:
                    match = re.match(r"\s+(.*)\s+", line)
                    if match:
                        return cls(line, LineType.PropValue, match)
                    else:
                        break
            else:
                if c == " ":
                    match = re.match(_resolution_mode_pattern, line)
                    if match:
                        return cls(line, LineType.ResolutionMode, match)
                    else:
                        break
                elif c == "\t":
                    tab_count += 1
            i += 1
        raise Exception(f"Line could not be categorized: '{line}'")


def _parse_screen(line: _Line) -> Screen:
    d = line.m.groupdict()

    screen: Screen = {"devices": []}  # type: ignore # Will be populated, but not immediately.
    for k, v in d.items():
        screen[k] = int(v)

    return screen


def _parse_device(line: _Line) -> Device:
    matches = line.m.groupdict()

    device: Device = {
        "props": defaultdict(list),
        "resolution_modes": [],
        "is_connected": matches["is_connected"] == "connected",
        "is_primary": matches["is_primary"] is not None
        and len(matches["is_primary"]) > 0,
        "device_name": matches["device_name"],
        "rotation": matches["rotation"] or "normal",
        "reflection": matches["reflection"] or "normal",
    }
    for k, v in matches.items():
        if k not in {
            "is_connected",
            "is_primary",
            "device_name",
            "rotation",
            "reflection",
        }:
            try:
                if v:
                    device[k] = int(v)
            except ValueError:
                raise Exception([f"{line.s} : {k} - {v} is not int-able"])

    return device


def _parse_resolution_mode(line: _Line) -> ResolutionMode:
    frequencies: List[Frequency] = []

    d = line.m.groupdict()
    resolution_width = int(d["resolution_width"])
    resolution_height = int(d["resolution_height"])
    is_high_resolution = d["is_high_resolution"] is not None

    mode: ResolutionMode = {
        "resolution_width": resolution_width,
        "resolution_height": resolution_height,
        "is_high_resolution": is_high_resolution,
        "frequencies": frequencies,
    }

    result = re.finditer(_frequencies_pattern, d["rest"])
    if not result:
        return mode

    for match in result:
        d = match.groupdict()
        frequency = float(d["frequency"])
        is_current = len(d["star"].strip()) > 0
        is_preferred = len(d["plus"].strip()) > 0
        f: Frequency = {
            "frequency": frequency,
            "is_current": is_current,
            "is_preferred": is_preferred,
        }
        mode["frequencies"].append(f)
    return mode


def _parse_props(index: int, line: _Line, lines: List[str]) -> Tuple[int, Props]:
    tmp_props: Dict[str, List[str]] = {}
    key = ""
    while index <= len(lines):
        if line.t == LineType.PropKey:
            d = line.m.groupdict()
            # See _prop_key_pattern
            key = d["key"]
            maybe_value = d["maybe_value"]
            if not maybe_value:
                tmp_props[key] = []
            else:
                tmp_props[key] = [maybe_value]
        elif line.t == LineType.PropValue:
            tmp_props[key].append(line.s.strip())
        else:
            # We've gone past our props and need to ascend
            index = index - 1
            break
        index += 1
        try:
            line = _Line.categorize(lines[index])
        except:
            pass

    props: Props = {}
    if "EDID" in tmp_props:
        edid = Edid(EdidHelper.hex2bytes("".join(tmp_props["EDID"])))
        model: EdidModel = {
            "name": edid.name or "Generic",
            "product_id": str(edid.product),
            "serial_number": str(edid.serial),
        }
        props["EdidModel"] = model

    return index, {**tmp_props, **props}


def parse(data: str, raw: bool = False, quiet: bool = False) -> Response:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        Dictionary. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    index = 0
    lines = data.splitlines()
    screen, device = None, None

    # temporary fix to ignore specific unhandled lines
    ignore_pattern = re.compile(r'^\s+(h|v):\s+(height|width)\s+\d+\s+start\s+\d+\s+end')

    result: Response = {"screens": []}
    if jc.utils.has_data(data):
        while index < len(lines):

            # temporary fix to ignore specific unhandled lines
            ignore_re = ignore_pattern.match(lines[index])
            if ignore_re:
                index += 1
                continue

            line = _Line.categorize(lines[index])
            if line.t == LineType.Screen:
                screen = _parse_screen(line)
                result["screens"].append(screen)
            elif line.t == LineType.Device:
                device = _parse_device(line)
                if not screen:
                    raise Exception("There should be an identifiable screen")
                screen["devices"].append(device)
            elif line.t == LineType.ResolutionMode:
                resolution_mode = _parse_resolution_mode(line)
                if not device:
                    raise Exception("Undefined device")
                device["resolution_modes"].append(resolution_mode)
            elif line.t == LineType.PropKey:
                # Props needs to be state aware, it owns the index.
                ix, props = _parse_props(index, line, lines)
                index = ix
                if not device:
                    raise Exception("Undefined device")
                device["props"] = props
            index += 1

    return result



import os
import json
import unittest
import jc.parsers.hciconfig

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class MyTests(unittest.TestCase):

    # input
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/hciconfig.out'), 'r', encoding='utf-8') as f:
        centos_7_7_hciconfig = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/hciconfig-a.out'), 'r', encoding='utf-8') as f:
        centos_7_7_hciconfig_a = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-20.04/hciconfig.out'), 'r', encoding='utf-8') as f:
        ubuntu_20_4_hciconfig = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-20.04/hciconfig-a.out'), 'r', encoding='utf-8') as f:
        ubuntu_20_4_hciconfig_a = f.read()

    # output
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/hciconfig.json'), 'r', encoding='utf-8') as f:
        centos_7_7_hciconfig_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/hciconfig-a.json'), 'r', encoding='utf-8') as f:
        centos_7_7_hciconfig_a_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-20.04/hciconfig.json'), 'r', encoding='utf-8') as f:
        ubuntu_20_4_hciconfig_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-20.04/hciconfig-a.json'), 'r', encoding='utf-8') as f:
        ubuntu_20_4_hciconfig_a_json = json.loads(f.read())


    def test_hciconfig_nodata(self):
        """
        Test plain 'hciconfig' with no data
        """
        self.assertEqual(jc.parsers.hciconfig.parse('', quiet=True), [])

    def test_hciconfig_centos_7_7(self):
        """
        Test plain 'hciconfig' on Centos 7.7
        """
        self.assertEqual(jc.parsers.hciconfig.parse(self.centos_7_7_hciconfig, quiet=True), self.centos_7_7_hciconfig_json)

    def test_hciconfig_a_centos_7_7(self):
        """
        Test 'hciconfig -a' on Centos 7.7
        """
        self.assertEqual(jc.parsers.hciconfig.parse(self.centos_7_7_hciconfig_a, quiet=True), self.centos_7_7_hciconfig_a_json)

    def test_hciconfig_ubuntu_18_4(self):
        """
        Test plain 'hciconfig' on Ubuntu 20.4
        """
        self.assertEqual(jc.parsers.hciconfig.parse(self.ubuntu_20_4_hciconfig, quiet=True), self.ubuntu_20_4_hciconfig_json)

    def test_hciconfig_a_ubuntu_18_4(self):
        """
        Test 'hciconfig -a' on Ubuntu 20.4
        """
        self.assertEqual(jc.parsers.hciconfig.parse(self.ubuntu_20_4_hciconfig_a, quiet=True), self.ubuntu_20_4_hciconfig_a_json)


if __name__ == '__main__':
    unittest.main()



"""jc - JSON test utils"""
import json
import os
import sys
from pathlib import Path
import jc

VERBOSE = False
if '-v' in sys.argv or '--verbose' in sys.argv:
    VERBOSE = True

def _test_print(data: str) -> None:
    if VERBOSE:
        print(data)

def _open_file(file_path, ext):
    return open(Path(file_path).with_suffix(ext), 'r', encoding='utf-8')


def _get_base_dir(file_path):
    return os.path.dirname(os.path.abspath(file_path))


def _get_parser_name_from_path(parser_path):
    return Path(parser_path).stem[len('test_'):]


def _get_fixtures(base_dir, parser_name):
    fixtures = {x.stem: str(x.with_suffix('')) for x in
                (list(Path(base_dir).glob(f"**/{parser_name}--*.*")))}
    return fixtures


def run_no_data(self, test_parser_path, expected):
    """Call this function to run a test for no input data for a parser"""
    parser_name = _get_parser_name_from_path(test_parser_path)

    with self.subTest(f"'no data test' for parser '{parser_name}': "):
        self.assertEqual(jc.parse(parser_name, '', quiet=True), expected)


def run_all_fixtures(self, test_parser_path):
    """Call this function to run tests for all fixtures for a parser"""
    parser_name = _get_parser_name_from_path(test_parser_path)
    base_dir = _get_base_dir(test_parser_path)
    fixtures = _get_fixtures(base_dir, parser_name).items()

    if not fixtures:
        raise ValueError(f"No fixtures found for '{parser_name}' tests!")

    _test_print(f"\n  Run all fixtures for parser '{parser_name}':")

    for file, file_path in fixtures:
        _test_print(f"  - test '{parser_name}' parser with fixture: '{file}'")

        with self.subTest(f"fixture: '{file}'"):
            with _open_file(file_path, '.out') as in_file, \
                    _open_file(file_path, '.json') as json_file:
                f_in = in_file.read()
                f_json = json.loads(json_file.read())

                self.assertEqual(
                    jc.parse(parser_name, f_in, quiet=True),
                    f_json,
                    f"Should be equal for test files: '{file_path}.*'"
                )



import os
import unittest
import json
import jc.parsers.pidstat
from jc.exceptions import ParseError

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class MyTests(unittest.TestCase):

    # input
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/pidstat.out'), 'r', encoding='utf-8') as f:
        centos_7_7_pidstat = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/pidstat-hl.out'), 'r', encoding='utf-8') as f:
        centos_7_7_pidstat_hl = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/pidstat-hdlrsuw.out'), 'r', encoding='utf-8') as f:
        centos_7_7_pidstat_hdlrsuw = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/pidstat-hdlrsuw-2-5.out'), 'r', encoding='utf-8') as f:
        centos_7_7_pidstat_hdlrsuw_2_5 = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/pidstat-ht.out'), 'r', encoding='utf-8') as f:
        generic_pidstat_ht = f.read()

    # output
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/pidstat-hl.json'), 'r', encoding='utf-8') as f:
        centos_7_7_pidstat_hl_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/pidstat-hdlrsuw.json'), 'r', encoding='utf-8') as f:
        centos_7_7_pidstat_hdlrsuw_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/pidstat-hdlrsuw-2-5.json'), 'r', encoding='utf-8') as f:
        centos_7_7_pidstat_hdlrsuw_2_5_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/pidstat-ht.json'), 'r', encoding='utf-8') as f:
        generic_pidstat_ht_json = json.loads(f.read())


    def test_pidstat_nodata(self):
        """
        Test 'pidstat' with no data
        """
        self.assertEqual(jc.parsers.pidstat.parse('', quiet=True), [])

    def test_pidstat(self):
        """
        Test 'pidstat' without -h... should raise ParseError
        """
        self.assertRaises(ParseError, jc.parsers.pidstat.parse, self.centos_7_7_pidstat, quiet=True)

    def test_pidstat_hl_centos_7_7(self):
        """
        Test 'pidstat -hl' on Centos 7.7
        """
        self.assertEqual(jc.parsers.pidstat.parse(self.centos_7_7_pidstat_hl, quiet=True), self.centos_7_7_pidstat_hl_json)

    def test_pidstat_hdlrsuw_centos_7_7(self):
        """
        Test 'pidstat -hdlrsuw' on Centos 7.7
        """
        self.assertEqual(jc.parsers.pidstat.parse(self.centos_7_7_pidstat_hdlrsuw, quiet=True), self.centos_7_7_pidstat_hdlrsuw_json)

    def test_pidstat_hdlrsuw_2_5_centos_7_7(self):
        """
        Test 'pidstat -hdlrsuw 2 5' on Centos 7.7
        """
        self.assertEqual(jc.parsers.pidstat.parse(self.centos_7_7_pidstat_hdlrsuw_2_5, quiet=True), self.centos_7_7_pidstat_hdlrsuw_2_5_json)

    def test_pidstat_ht(self):
        """
        Test 'pidstat -hT'
        """
        self.assertEqual(jc.parsers.pidstat.parse(self.generic_pidstat_ht, quiet=True), self.generic_pidstat_ht_json)


if __name__ == '__main__':
    unittest.main()



import unittest
import json
import jc.parsers.email_address


class MyTests(unittest.TestCase):
    def test_email_address_nodata(self):
        """
        Test 'email_address' with no data
        """
        self.assertEqual(jc.parsers.email_address.parse('', quiet=True), {})


    def test_simple_email(self):
        """
        Test simple email address
        """
        data = r'fred@example.com'
        expected = json.loads(r'''{"username":"fred","domain":"example.com","local":"fred","local_plus_suffix":null}''')
        self.assertEqual(jc.parsers.email_address.parse(data, quiet=True), expected)


    def test_plus_email(self):
        """
        Test email address with plus syntax
        """
        data = r'fred+spam@example.com'
        expected = json.loads(r'''{"username":"fred","domain":"example.com","local":"fred+spam","local_plus_suffix":"spam"}''')
        self.assertEqual(jc.parsers.email_address.parse(data, quiet=True), expected)


if __name__ == '__main__':
    unittest.main()



r"""jc - JSON Convert `typeset` and `declare` Bash internal command output parser

Convert `typeset` and `declare` bash internal commands with no options or the
following:  `-a`, `-A`, `-i`, `-l`, `-p`, `-r`, `-u`, and `-x`

Note: function parsing is not supported (e.g. `-f` or `-F`)

Usage (cli):

    $ typeset | jc --typeset

Usage (module):

    import jc
    result = jc.parse('typeset', typeset_command_output)

Schema:

    [
      {
        "name":         string,
        "value":        string/integer/array/object/null,    # [0]
        "type":         string,                              # [1]
        "readonly":     boolean/null,
        "integer":      boolean/null,
        "lowercase":    boolean/null,
        "uppercase":    boolean/null,
        "exported":     boolean/null
      }
    ]

    Key/value pairs other than `name`, `value`, and `type` will only be non-null
    when the information is available from the `typeset` or `declare` output.

    If declare options are not given to `jc` within the `typeset` output, then
    it will assume all arrays are simple `array` type.

    [0] Based on type. `variable` type is null if not set, a string when the
        bash variable is set unless the `integer` field is set to `True`, then
        the type is integer. `array` type is an array of strings or integers as
        above. `associative` type is an object of key/value pairs where values
        are strings or integers as above. Objects have the schema of:

        {
          "<key1>": string/integer,
          "<key2>": string/integer
        }

    [1] Possible values: `variable`, `array`, or `associative`

Examples:

    $ typeset -p | jc --typeset -p
    [
      {
        "name": "associative_array",
        "value": {
          "key2": "abc",
          "key3": "1 2 3",
          "key1": "hello \"world\""
        },
        "type": "associative",
        "readonly": false,
        "integer": false,
        "lowercase": false,
        "uppercase": false,
        "exported": false
      },
      {
        "name": "integers_associative_array",
        "value": {
          "one": 1,
          "two": 500,
          "three": 999
        },
        "type": "associative",
        "readonly": false,
        "integer": true,
        "lowercase": false,
        "uppercase": false,
        "exported": false
      }
    ]

    $ typeset -p | jc --typeset -p -r
    [
      {
        "name": "associative_array",
        "value": {
          "key2": "abc",
          "key3": "1 2 3",
          "key1": "hello \"world\""
        },
        "type": "associative",
        "readonly": false,
        "integer": false,
        "lowercase": false,
        "uppercase": false,
        "exported": false
      },
      {
        "name": "integers_associative_array",
        "value": {
          "one": "1",
          "two": "500",
          "three": "999"
        },
        "type": "associative",
        "readonly": false,
        "integer": true,
        "lowercase": false,
        "uppercase": false,
        "exported": false
      }
    ]
"""
import shlex
import re
from typing import List, Dict
from jc.jc_types import JSONDictType
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.0'
    description = '`typeset` and `declare` command parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux', 'darwin', 'cygwin', 'win32', 'aix', 'freebsd']
    tags = ['command']


__version__ = info.version

VAR_DEF_PATTERN = re.compile(r'(?P<name>[a-zA-Z_][a-zA-Z0-9_]*)=(?P<val>[^(][^[].+)$')
SIMPLE_ARRAY_DEF_PATTERN = re.compile(r'(?P<name>[a-zA-Z_][a-zA-Z0-9_]*)=(?P<body>\(\[\d+\]=.+\))$')
ASSOCIATIVE_ARRAY_DEF_PATTERN = re.compile(r'(?P<name>[a-zA-Z_][a-zA-Z0-9_]*)=(?P<body>\(\[[a-zA-Z_][a-zA-Z0-9_]*\]=.+\))$')
EMPTY_ARRAY_DEF_PATTERN = re.compile(r'(?P<name>[a-zA-Z_][a-zA-Z0-9_]*)=\(\)$')
EMPTY_VAR_DEF_PATTERN = re.compile(r'declare\s.+\s(?P<name>[a-zA-Z_][a-zA-Z0-9_]*)$')
DECLARE_OPTS_PATTERN = re.compile(r'declare\s(?P<options>.+?)\s[a-zA-Z_][a-zA-Z0-9_]*')


def _process(proc_data: List[JSONDictType]) -> List[JSONDictType]:
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        List of Dictionaries. Structured to conform to the schema.
    """
    for item in proc_data:
        if item['type'] == 'variable' and item['integer']:
            item['value'] = jc.utils.convert_to_int(item['value'])

        elif item['type'] == 'array' and item['integer'] \
            and isinstance(item['value'], list):

            new_num_list = []
            for number in item['value']:
                new_num_list.append(jc.utils.convert_to_int(number))

            item['value'] = new_num_list

        elif (item['type'] == 'array' and item['integer'] \
            and isinstance(item['value'], dict)) \
            or (item['type'] == 'associative' and item['integer']):

            new_num_dict: Dict[str, int] = {}
            for key, val in item['value'].items():
                new_num_dict.update({key: jc.utils.convert_to_int(val)})

            item['value'] = new_num_dict

    return proc_data


def _get_simple_array_vals(body: str) -> List[str]:
    body = _remove_bookends(body)
    body_split = shlex.split(body)
    values = []
    for item in body_split:
        _, val = item.split('=', maxsplit=1)
        values.append(_remove_quotes(val))
    return values


def _get_associative_array_vals(body: str) -> Dict[str, str]:
    body = _remove_bookends(body)
    body_split = shlex.split(body)
    values: Dict = {}
    for item in body_split:
        key, val = item.split('=', maxsplit=1)
        key = _remove_bookends(key, '[', ']')
        values.update({key: val})
    return values


def _get_declare_options(line: str, type_hint: str = 'variable') -> Dict:
    opts = {
        'type': type_hint,
        'readonly': None,
        'integer':  None,
        'lowercase': None,
        'uppercase': None,
        'exported': None
    }

    opts_map = {
        'r': 'readonly',
        'i': 'integer',
        'l': 'lowercase',
        'u': 'uppercase',
        'x': 'exported'
    }

    declare_opts_match = re.match(DECLARE_OPTS_PATTERN, line)
    if declare_opts_match:
        for opt in declare_opts_match['options']:
            if opt in opts_map:
                opts[opts_map[opt]] = True
                continue
        if 'a' in declare_opts_match['options']:
            opts['type'] = 'array'
        elif 'A' in declare_opts_match['options']:
            opts['type'] = 'associative'

        # flip all remaining Nones to False
        for option in opts.items():
            key, val = option
            if val is None:
                opts[key] = False
    return opts


def _remove_bookends(data: str, start_char: str = '(', end_char: str = ')') -> str:
    if data.startswith(start_char) and data.endswith(end_char):
        return data[1:-1]
    return data


def _remove_quotes(data: str, remove_char: str ='"') -> str:
    if data.startswith(remove_char) and data.endswith(remove_char):
        return data[1:-1]
    return data


def parse(
    data: str,
    raw: bool = False,
    quiet: bool = False
) -> List[JSONDictType]:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output: List[Dict] = []

    if jc.utils.has_data(data):

        for line in filter(None, data.splitlines()):

            item = {
                "name": '',
                "value": '',
                "type": None,
                "readonly": None,
                "integer":  None,
                "lowercase": None,
                "uppercase": None,
                "exported": None
            }

            # regular variable
            var_def_match = re.search(VAR_DEF_PATTERN, line)
            if var_def_match:
                item['name'] = var_def_match['name']
                item['value'] = _remove_quotes(var_def_match['val'])
                item.update(_get_declare_options(line, 'variable'))
                raw_output.append(item)
                continue

            # empty variable
            empty_var_def_match = re.search(EMPTY_VAR_DEF_PATTERN, line)
            if empty_var_def_match:
                item['name'] = empty_var_def_match['name']
                item['value'] = None
                item.update(_get_declare_options(line, 'variable'))
                raw_output.append(item)
                continue

            # simple array
            simple_arr_def_match = re.search(SIMPLE_ARRAY_DEF_PATTERN, line)
            if simple_arr_def_match:
                item['name'] = simple_arr_def_match['name']
                item['value'] = _get_simple_array_vals(simple_arr_def_match['body'])
                item.update(_get_declare_options(line, 'array'))
                raw_output.append(item)
                continue

            # associative array
            associative_arr_def_match = re.search(ASSOCIATIVE_ARRAY_DEF_PATTERN, line)
            if associative_arr_def_match:
                item['name'] = associative_arr_def_match['name']
                item['value'] = _get_associative_array_vals(associative_arr_def_match['body'])
                item.update(_get_declare_options(line, 'associative'))
                raw_output.append(item)
                continue

            # empty array
            empty_arr_def_match = re.search(EMPTY_ARRAY_DEF_PATTERN, line)
            if empty_arr_def_match:
                item['name'] = empty_arr_def_match['name']
                item['value'] = []
                item.update(_get_declare_options(line, 'array'))
                raw_output.append(item)
                continue

    return raw_output if raw else _process(raw_output)



r"""jc - JSON Convert Email Address string parser

Usage (cli):

    $ echo "username@example.com" | jc --email-address

Usage (module):

    import jc
    result = jc.parse('email_address', email_address_string)

Schema:

    {
      "username":             string,
      "domain":               string,
      "local":                string,
      "local_plus_suffix":    string or null
    }

Examples:

    $ echo 'joe.user@gmail.com' | jc --email-address -p
    {
      "username": "joe.user",
      "domain": "gmail.com",
      "local": "joe.user",
      "local_plus_suffix": null
    }

    $ echo 'joe.user+spam@gmail.com' | jc --email-address -p
    {
      "username": "joe.user",
      "domain": "gmail.com",
      "local": "joe.user+spam",
      "local_plus_suffix": "spam"
    }
"""
from typing import Dict
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.1'
    description = 'Email Address string parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux', 'darwin', 'cygwin', 'win32', 'aix', 'freebsd']
    tags = ['standard', 'string', 'slurpable']


__version__ = info.version


def _process(proc_data: Dict) -> Dict:
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (Dictionary) raw structured data to process

    Returns:

        Dictionary. Structured to conform to the schema.
    """
    return proc_data


def parse(
    data: str,
    raw: bool = False,
    quiet: bool = False
) -> Dict:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        Dictionary. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output: Dict = {}

    if jc.utils.has_data(data):
        data = data.strip()
        domain = data.rsplit('@', maxsplit=1)[1]
        local = data.rsplit('@', maxsplit=1)[0]

        local_plus_prefix = None
        local_plus_suffix = None
        try:
            local_plus_suffix = local.split('+', maxsplit=1)[1]
            local_plus_prefix = local.split('+', maxsplit=1)[0]
        except IndexError:
            pass

        raw_output = {
            'username': local_plus_prefix or local,
            'domain': domain,
            'local': local,
            'local_plus_suffix': local_plus_suffix
        }

    return raw_output if raw else _process(raw_output)



r"""jc - JSON Convert `finger` command output parser

Supports `-s` output option. Does not support the `-l` detail option.

Usage (cli):

    $ finger | jc --finger

or

    $ jc finger

Usage (module):

    import jc
    result = jc.parse('finger', finger_command_output)

Schema:

    [
      {
        "login":                string,
        "name":                 string,
        "tty":                  string,
        "idle":                 string,     # null if empty
        "login_time":           string,
        "details":              string,
        "tty_writeable":        boolean,
        "idle_minutes":         integer,
        "idle_hours":           integer,
        "idle_days":            integer,
        "total_idle_minutes":   integer
      }
    ]

Examples:

    $ finger | jc --finger -p
    [
      {
        "login": "jdoe",
        "name": "John Doe",
        "tty": "tty1",
        "idle": "14d",
        "login_time": "Mar 22 21:14",
        "tty_writeable": false,
        "idle_minutes": 0,
        "idle_hours": 0,
        "idle_days": 14,
        "total_idle_minutes": 20160
      },
      {
        "login": "jdoe",
        "name": "John Doe",
        "tty": "pts/0",
        "idle": null,
        "login_time": "Apr  5 15:33",
        "details": "(192.168.1.22)",
        "tty_writeable": true,
        "idle_minutes": 0,
        "idle_hours": 0,
        "idle_days": 0,
        "total_idle_minutes": 0
      },
      ...
    ]

    $ finger | jc --finger -p -r
    [
      {
        "login": "jdoe",
        "name": "John Doe",
        "tty": "*tty1",
        "idle": "14d",
        "login_time": "Mar 22 21:14"
      },
      {
        "login": "jdoe",
        "name": "John Doe",
        "tty": "pts/0",
        "idle": null,
        "login_time": "Apr  5 15:33",
        "details": "(192.168.1.22)"
      },
      ...
    ]
"""
import re
import jc.utils
import jc.parsers.universal


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.2'
    description = '`finger` command parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux', 'darwin', 'cygwin', 'freebsd']
    magic_commands = ['finger']
    tags = ['command']


__version__ = info.version


def _process(proc_data):
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        List of Dictionaries. Structured data to conform to the schema.
    """
    for entry in proc_data:
        if 'tty' in entry:
            entry['tty_writeable'] = True
            if '*' in entry['tty']:
                entry['tty'] = entry['tty'].replace('*', '')
                entry['tty_writeable'] = False

        if 'idle' in entry:
            entry['idle_minutes'] = 0
            entry['idle_hours'] = 0
            entry['idle_days'] = 0

            if entry['idle'] == '-':
                entry['idle'] = None

            if entry['idle'] and entry['idle'].isnumeric():
                entry['idle_minutes'] = int(entry['idle'])

            if entry['idle'] and ':' in entry['idle']:
                entry['idle_hours'] = int(entry['idle'].split(':')[0])
                entry['idle_minutes'] = int(entry['idle'].split(':')[1])

            if entry['idle'] and 'd' in entry['idle']:
                entry['idle_days'] = int(entry['idle'].replace('d', ''))

            entry['total_idle_minutes'] = (entry['idle_days'] * 1440) + \
                                          (entry['idle_hours'] * 60) + \
                                          entry['idle_minutes']

        if 'details' in entry:
            if not entry['details']:
                del entry['details']

    return proc_data


def parse(data, raw=False, quiet=False):
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output = []

    if jc.utils.has_data(data):
        # Finger output is an abomination that is nearly unparsable. But there
        # is a way:
        # First find the location of the last character of 'Idle' in the table
        # and cut all lines at that spot. Data before that spot can use the
        # unviversal.sparse_table_parse function. All data after that spot can
        # be run through regex to find the login datetime and possibly other fields.

        data_lines = list(filter(None, data.splitlines()))
        sep_col = data_lines[0].find('Idle') + 4
        first_half = []
        second_half = []

        for line in data_lines:
            first_half.append(line[:sep_col])
            second_half.append(line[sep_col:])

        first_half[0] = first_half[0].lower()

        # parse the first half
        raw_output =  jc.parsers.universal.sparse_table_parse(first_half)

        # use regex to get login datetime and 'other' data
        pattern = re.compile(r'([A-Z][a-z]{2}\s+\d{1,2}\s+)(\d\d:\d\d|\d{4})(\s?.+)?$')

        # remove header row from list
        second_half.pop(0)

        for index, line in enumerate(second_half):
            dt = re.search(pattern, line)
            if dt:
                if dt.group(1) and dt.group(2):
                    raw_output[index]['login_time'] = dt.group(1).strip() + ' ' + dt.group(2).strip()
                if dt.group(3):
                    raw_output[index]['details'] = dt.group(3).strip()

    if raw:
        return raw_output
    else:
        return _process(raw_output)



# Copyright (c) 2016, Samantha Marshall (http://pewpewthespells.com)
# All rights reserved.
#
# https://github.com/samdmarshall/pbPlist
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# 3. Neither the name of Samantha Marshall nor the names of its contributors may
# be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.

from . import StrParse

def PushIndent(indent_level):
    return indent_level + 1

def PopIndent(indent_level):
    return indent_level - 1

def WriteIndent(level=0):
    output = ''
    for _index in range(level):
        output += '\t'
    return output

def WriteNewline(level=0, indent=True):
    output = '\n'
    if indent:
        output += WriteIndent(level)
    return output

class pbItem(object):
    def __init__(self, value=None, type_name=None, annotation=None):
        if value != None and type_name != None:
            self.value = value
            if type_name not in KnownTypes.keys(): # pragma: no cover
                message = 'Unknown type "'+type_name+'" passed to '+self.__class__.__name__+' initializer!'
                raise TypeError(message)
            self.type_name = type_name
            self.annotation = annotation
        else: # pragma: no cover
            message = 'The class "'+self.__class__.__name__+'" must be initialized with a non-None value'
            raise ValueError(message)

    def __eq__(self, other):
        is_equal = False
        if isinstance(other, pbItem):
            other = other.value
        if type(other) is type(self.value):
            is_equal = self.value.__eq__(other)
        return is_equal

    def __hash__(self):
        return self.value.__hash__()

    def __repr__(self):
        return self.value.__repr__()

    def __iter__(self):
        return self.value.__iter__()

    def __getattr__(self, attrib):
        return self.value.__getattr__(attrib)

    def __str__(self):
        return self.writeStringRep(0, False)[0]

    def __getitem__(self, key):
        return self.value.__getitem__(key)

    def __setitem__(self, key, value):
        self.value.__setitem__(key, value)

    def __len__(self):
        return self.value.__len__()

    def __contains__(self, item):
        return self.value.__contains__(item)

    def __get__(self, obj, objtype):
        return self.value.__get__(obj, objtype)

    def writeStringRep(self, indent_level=0, pretty=True):
        return self.writeString(indent_level, pretty)

    def writeString(self, indent_level=0, pretty=True): # pylint: disable=no-self-use,unused-variable,unused-argument ; # pragma: no cover
        message = 'This is a base class, it cannot write!'
        raise Exception(message)

    def nativeType(self):
        return self.value

    def writeAnnotation(self):
        output_string = ''
        if self.annotation != None and len(self.annotation) > 0:
            output_string += ' '
            output_string += '/*'
            output_string += self.annotation
            output_string += '*/'
        return output_string

class pbString(pbItem):
    def writeString(self, indent_level=0, pretty=True):
        string_string = ''
        string_string += self.value
        if pretty is True:
            string_string += self.writeAnnotation()
        return (string_string, indent_level)

class pbQString(pbItem):
    def writeStringRep(self, indent_level=0, pretty=True):
        qstring_string = ''
        for character in self.value:
            qstring_string += StrParse.SanitizeCharacter(character)
        return (qstring_string, indent_level)

    def writeString(self, indent_level=0, pretty=True):
        qstring_string = ''
        qstring_string += '"'
        string_rep, indent_level = self.writeStringRep(indent_level, pretty)
        qstring_string += string_rep
        qstring_string += '"'
        if pretty is True:
            qstring_string += self.writeAnnotation()
        return (qstring_string, indent_level)

class pbData(pbItem):
    def writeString(self, indent_level=0, pretty=True):
        data_string = ''
        indent_level = PushIndent(indent_level)
        data_string += '<'
        grouping_byte_counter = 0
        grouping_line_counter = 0
        for hex_byte in map(ord, self.value.decode()):
            data_string += format(hex_byte, 'x')
            grouping_byte_counter += 1
            # write a space every 4th byte
            if grouping_byte_counter == 4:
                data_string += ' '
                grouping_byte_counter = 0
                grouping_line_counter += 1
            # write a newline every 4th grouping of bytes
            if grouping_line_counter == 4:
                data_string += WriteNewline(indent_level)
                data_string += ' ' # indent an additional space to make the byte groupings line up
                grouping_line_counter = 0
        data_string += '>'
        if pretty is True:
            data_string += self.writeAnnotation()
        indent_level = PopIndent(indent_level)
        return (data_string, indent_level)

class pbDictionary(pbItem):
    def nativeType(self):
        new_value = dict()
        for key in self.keys():
            value = self[key]
            new_value[str(key)] = value.nativeType()
        return new_value
    def writeString(self, indent_level=0, pretty=True):
        dictionary_string = ''
        dictionary_string += '{'
        has_sorted_keys, keys_array = self.value.sortedKeys()
        dictionary_string += WriteNewline(indent_level, not has_sorted_keys)
        indent_level = PushIndent(indent_level)
        previous_value_type = None
        if len(keys_array) == 0:
            indent_level = PopIndent(indent_level)
        else:
            if not has_sorted_keys:
                dictionary_string += '\t'
            for key in keys_array:
                if has_sorted_keys:
                    current_value_type = str(self.value[key]['isa'])
                    if previous_value_type != current_value_type:
                        if previous_value_type != None:
                            dictionary_string += '/* End '+previous_value_type+' section */'
                            dictionary_string += WriteNewline(indent_level, False)
                        previous_value_type = current_value_type
                        dictionary_string += '\n/* Begin '+current_value_type+' section */'
                        dictionary_string += WriteNewline(indent_level)
                    else:
                        dictionary_string += WriteIndent(indent_level)
                write_string, indent_level = key.writeString(indent_level, pretty)
                dictionary_string += write_string
                dictionary_string += ' = '
                write_string, indent_level = self.value[key].writeString(indent_level, pretty)
                dictionary_string += write_string
                dictionary_string += ';'
                should_indent = True
                is_last_key = (key == keys_array[-1])
                if is_last_key:
                    if has_sorted_keys:
                        dictionary_string += WriteNewline(indent_level, False)
                        dictionary_string += '/* End '+previous_value_type+' section */'
                    indent_level = PopIndent(indent_level)
                else:
                    if has_sorted_keys:
                        should_indent = False
                dictionary_string += WriteNewline(indent_level, should_indent)
        dictionary_string += '}'
        return (dictionary_string, indent_level)

class pbArray(pbItem):
    def nativeType(self):
        new_value = [item.nativeType() for item in self.value]
        return new_value
    def writeString(self, indent_level=0, pretty=True):
        array_string = ''
        array_string += '('
        array_string += WriteNewline(indent_level)
        indent_level = PushIndent(indent_level)
        values_array = list(self.value)
        if len(values_array) == 0:
            indent_level = PopIndent(indent_level)
        else:
            array_string += '\t'
            for value in values_array:
                write_string, indent_level = value.writeString(indent_level, pretty)
                array_string += write_string
                if value != values_array[-1]:
                    array_string += ','
                else:
                    indent_level = PopIndent(indent_level)
                array_string += WriteNewline(indent_level)
        array_string += ')'
        return (array_string, indent_level)

KnownTypes = {
    'string': pbString,
    'qstring': pbQString,
    'data': pbData,
    'dictionary': pbDictionary,
    'array': pbArray,
}

def pbItemResolver(obj, type_name):
    initializer = KnownTypes[type_name]
    if initializer:
        return initializer(obj, type_name)
    else: # pragma: no cover
        message = 'Unknown type "'+type_name+'" passed to pbItemResolver!'
        raise TypeError(message)



import os
import unittest
from jc.parsers.os_prober import parse

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class MyTests(unittest.TestCase):

    def test_os_prober_nodata(self):
        """
        Test 'os_prober' with no data
        """
        self.assertEqual(parse('', quiet=True), {})

    def test_os_prober_1(self):
        """
        Test 'os_prober' 1
        """
        self.assertEqual(
            parse('/dev/sda1:Windows 7 (loader):Windows:chain', quiet=True),
            {"partition":"/dev/sda1","name":"Windows 7 (loader)","short_name":"Windows","type":"chain"}
        )

    def test_os_prober_2(self):
        """
        Test 'os_prober' 2
        """
        self.assertEqual(
            parse('/dev/sda1:Windows 10:Windows:chain', quiet=True),
            {"partition":"/dev/sda1","name":"Windows 10","short_name":"Windows","type":"chain"}
        )

    def test_os_prober_3(self):
        """
        Test 'os_prober' 3
        """
        self.assertEqual(
            parse('/dev/sda1@/efi/Microsoft/Boot/bootmgfw.efi:Windows Boot Manager:Windows:efi', quiet=True),
            {"partition":"/dev/sda1","efi_bootmgr":"/efi/Microsoft/Boot/bootmgfw.efi","name":"Windows Boot Manager","short_name":"Windows","type":"efi"}
        )

    def test_os_prober_3_raw(self):
        """
        Test 'os_prober' 3 with raw output
        """
        self.assertEqual(
            parse('/dev/sda1@/efi/Microsoft/Boot/bootmgfw.efi:Windows Boot Manager:Windows:efi', quiet=True, raw=True),
            {"partition":"/dev/sda1@/efi/Microsoft/Boot/bootmgfw.efi","name":"Windows Boot Manager","short_name":"Windows","type":"efi"}
        )


if __name__ == '__main__':
    unittest.main()



r"""jc - JSON Convert disabled parser

This parser has been disabled due to an error in the parser code.
"""
from jc.exceptions import ParseError

class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.0'
    description = 'Disabled parser'
    author = 'N/A'
    author_email = 'N/A'
    compatible = ['linux', 'darwin', 'cygwin', 'win32', 'aix', 'freebsd']
    hidden = True


__version__ = info.version


def parse(
    data: str,
    raw: bool = False,
    quiet: bool = False
) -> dict:
    """Main text parsing function"""
    raise ParseError('This parser is disabled.')



r"""jc - JSON Convert `host` command output parser

Supports parsing of the most commonly used RR types (A, AAAA, MX, TXT)

Usage (cli):

    $ host google.com | jc --host

or

    $ jc host google.com

Usage (module):

    import jc
    result = jc.parse('host', host_command_output)

Schema:

    [
      {
        "hostname":     string,
        "address": [
                        string
        ],
        "v6-address": [
                        string
        ],
        "mail": [
                        string
        ]
      }
    ]

    [
      {
        "nameserver":   string,
        "zone":         string,
        "mname":        string,
        "rname":        string,
        "serial":       integer,
        "refresh":      integer,
        "retry":        integer,
        "expire":       integer,
        "minimum":      integer
      }
    ]

Examples:

    $ host google.com | jc --host
    [
      {
        "hostname": "google.com",
        "address": [
          "142.251.39.110"
        ],
        "v6-address": [
          "2a00:1450:400e:811::200e"
        ],
        "mail": [
          "smtp.google.com."
        ]
      }
    ]

    $ jc host -C sunet.se
    [
      {
        "nameserver": "2001:6b0:7::2",
        "zone": "sunet.se",
        "mname": "sunic.sunet.se.",
        "rname": "hostmaster.sunet.se.",
        "serial": "2023090401",
        "refresh": "28800",
        "retry": "7200",
        "expire": "604800",
        "minimum": "300"
      },
      {
        ...
      }
    ]
"""
from typing import Dict, List
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.0'
    description = '`host` command parser'
    author = 'Pettai'
    author_email = 'pettai@sunet.se'
    compatible = ['linux', 'darwin', 'cygwin', 'win32', 'aix', 'freebsd']
    tags = ['command']
    magic_commands = ['host']


__version__ = info.version


def _process(proc_data):
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        List of Dictionaries. Structured to conform to the schema.
    """

    int_list = {'serial', 'refresh', 'retry', 'expire', 'minimum'}

    for entry in proc_data:
        for key in entry:
            if key in int_list:
                entry[key] = jc.utils.convert_to_int(entry[key])

    return proc_data


def parse(data: str, raw: bool = False, quiet: bool = False):
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output: List[Dict] = []

    warned = False

    if jc.utils.has_data(data):

        addresses = []
        v6addresses = []
        mail = []
        text = []
        rrdata = {}
        soaparse = False

        for line in filter(None, data.splitlines()):
            line = line.strip()

            # default
            if ' has address ' in line:
                linedata = line.split(' ', maxsplit=3)
                hostname = linedata[0]
                address = linedata[3]
                addresses.append(address)
                rrdata.update({'hostname': hostname})
                rrdata.update({'address': addresses})
                continue

            if ' has IPv6 address ' in line:
                linedata = line.split(' ', maxsplit=4)
                hostname = linedata[0]
                v6address = linedata[4]
                v6addresses.append(v6address)
                rrdata.update({'hostname': hostname})
                rrdata.update({'v6-address': v6addresses})
                continue

            if ' mail is handled by ' in line:
                linedata = line.split(' ', maxsplit=6)
                hostname = linedata[0]
                mx = linedata[6]
                mail.append(mx)
                rrdata.update({'hostname': hostname})
                rrdata.update({'mail': mail})
                continue


            # TXT parsing
            if ' descriptive text ' in line:
                linedata = line.split('descriptive text "', maxsplit=1)
                hostname = linedata[0]
                txt = linedata[1].strip('"')
                text.append(txt)
                rrdata.update({'hostname': hostname})
                rrdata.update({'text': text})
                continue


            # -C / SOA parsing
            if line.startswith('Nameserver '):
                soaparse = True
                rrdata = {}
                linedata = line.split(' ', maxsplit=1)
                nameserverip = linedata[1].rstrip(':')
                rrdata.update({'nameserver': nameserverip})
                continue

            if ' has SOA record ' in line:
                linedata = line.split(' ', maxsplit=10)

                zone = linedata[0]
                mname = linedata[4]
                rname = linedata[5]
                serial = linedata[6]
                refresh = linedata[7]
                retry = linedata[8]
                expire = linedata[9]
                minimum = linedata[10]

                try:
                    rrdata.update(
                        {
                            'zone': zone,
                            'mname': mname,
                            'rname': rname,
                            'serial': serial,
                            'refresh': refresh,
                            'retry': retry,
                            'expire': expire,
                            'minimum': minimum 
                        },
                    )
                    raw_output.append(rrdata)

                except IndexError:
                    if not warned:
                        jc.utils.warning_message(['Unknown format detected.'])
                        warned = True

        if not soaparse:
            raw_output.append(rrdata)

    return raw_output if raw else _process(raw_output)



r"""jc - JSON Convert `bluetoothctl` command output parser

Supports the following `bluetoothctl` subcommands:
- `bluetoothctl list`
- `bluetoothctl show`
- `bluetoothctl show <ctrl>`
- `bluetoothctl devices`
- `bluetoothctl info <dev>`

Usage (cli):

    $ bluetoothctl info <dev> | jc --bluetoothctl
or

    $ jc bluetoothctl info <dev>

Usage (module):

    import jc
    result = jc.parse('bluetoothctl', bluetoothctl_command_output)

Schema:

Because bluetoothctl is handling two main entities, controllers and devices,
the schema is shared between them. Most of the fields are common between
a controller and a device but there might be fields corresponding to one entity.

    Controller:
    [
        {
            "manufacturer":         string,
            "version":              string,
            "name":                 string,
            "is_default":           boolean,
            "is_public":            boolean,
            "is_random":            boolean,
            "address":              string,
            "alias":                string,
            "class":                string,
            "powered":              string,
            "discoverable":         string,
            "discoverable_timeout": string,
            "pairable":             string,
            "modalias":             string,
            "discovering":          string,
            "uuids":                array
        }
    ]

    Device:
    [
        {
            "name":                 string,
            "is_public":            boolean,
            "is_random":            boolean,
            "address":              string,
            "alias":                string,
            "appearance":           string,
            "class":                string,
            "icon":                 string,
            "paired":               string,
            "bonded":               string,
            "trusted":              string,
            "blocked":              string,
            "connected":            string,
            "legacy_pairing":       string,
            "cable_pairing":        string,
            "rssi":                 int,
            "txpower":              int,
            "uuids":                array,
            "modalias":             string,
            "battery_percentage":   int
        }
    ]

Examples:

    $ bluetoothctl info EB:06:EF:62:B3:19 | jc --bluetoothctl -p
    [
        {
            "address": "22:06:33:62:B3:19",
            "is_public": true,
            "name": "TaoTronics TT-BH336",
            "alias": "TaoTronics TT-BH336",
            "class": "0x00240455",
            "icon": "audio-headset",
            "paired": "no",
            "bonded": "no",
            "trusted": "no",
            "blocked": "no",
            "connected": "no",
            "legacy_pairing": "no",
            "uuids": [
                "Advanced Audio Distribu.. (0000120d-0000-1000-8000-00805f9b34fb)",
                "Audio Sink                (0000130b-0000-1000-8000-00805f9b34fb)",
                "A/V Remote Control        (0000140e-0000-1000-8000-00805f9b34fb)",
                "A/V Remote Control Cont.. (0000150f-0000-1000-8000-00805f9b34fb)",
                "Handsfree                 (0000161e-0000-1000-8000-00805f9b34fb)",
                "Headset                   (00001708-0000-1000-8000-00805f9b34fb)",
                "Headset HS                (00001831-0000-1000-8000-00805f9b34fb)"
            ],
            "rssi": -52,
            "txpower": 4,
            "battery_percentage": 70
        }
    ]
"""
import re
from typing import List, Dict, Optional, Any
from jc.jc_types import JSONDictType
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.5'
    description = '`bluetoothctl` command parser'
    author = 'Jake Ob'
    author_email = 'iakopap at gmail.com'
    compatible = ['linux']
    magic_commands = ['bluetoothctl']
    tags = ['command']


__version__ = info.version

try:
    from typing import TypedDict

    Controller = TypedDict(
        "Controller",
        {
            "manufacturer": str,
            "version": str,
            "name": str,
            "is_default": bool,
            "is_public": bool,
            "is_random": bool,
            "address": str,
            "alias": str,
            "class": str,
            "powered": str,
            "power_state": str,
            "discoverable": str,
            "discoverable_timeout": str,
            "pairable": str,
            "modalias": str,
            "discovering": str,
            "uuids": List[str],
        },
    )
    Device = TypedDict(
        "Device",
        {
            "name": str,
            "is_public": bool,
            "is_random": bool,
            "address": str,
            "alias": str,
            "appearance": str,
            "class": str,
            "icon": str,
            "paired": str,
            "bonded": str,
            "trusted": str,
            "blocked": str,
            "connected": str,
            "legacy_pairing": str,
            "cable_pairing": str,
            "rssi": int,
            "txpower": int,
            "uuids": List[str],
            "modalias": str,
            "battery_percentage": int
        },
    )
except ImportError:
    Controller = Dict[str, Any]  # type: ignore
    Device = Dict[str, Any]  # type: ignore


_controller_head_pattern = r"Controller (?P<address>([0-9A-F]{2}:){5}[0-9A-F]{2}) (?P<name>.+)"

_controller_line_pattern = (
    r"(\s*Manufacturer:\s*(?P<manufacturer>.+)"
    + r"|\s*Version:\s*(?P<version>.+)"
    + r"|\s*Name:\s*(?P<name>.+)"
    + r"|\s*Alias:\s*(?P<alias>.+)"
    + r"|\s*Class:\s*(?P<class>.+)"
    + r"|\s*Powered:\s*(?P<powered>.+)"
    + r"|\s*PowerState:\s*(?P<power_state>.+)"
    + r"|\s*Discoverable:\s*(?P<discoverable>.+)"
    + r"|\s*DiscoverableTimeout:\s*(?P<discoverable_timeout>.+)"
    + r"|\s*Pairable:\s*(?P<pairable>.+)"
    + r"|\s*Modalias:\s*(?P<modalias>.+)"
    + r"|\s*Discovering:\s*(?P<discovering>.+)"
    + r"|\s*UUID:\s*(?P<uuid>.+))"
)

def _parse_controller(next_lines: List[str]) -> Optional[Controller]:
    next_line = next_lines.pop()
    result = re.match(_controller_head_pattern, next_line)

    if not result:
        next_lines.append(next_line)
        return None

    matches = result.groupdict()

    name = matches["name"]

    if name.endswith("not available"):
        return None

    controller: Controller = {
            "manufacturer": '',
            "version": '',
            "name": '',
            "is_default": False,
            "is_public": False,
            "is_random": False,
            "address": matches["address"],
            "alias": '',
            "class": '',
            "powered": '',
            "power_state": '',
            "discoverable": '',
            "discoverable_timeout": '',
            "pairable": '',
            "modalias": '',
            "discovering": '',
            "uuids": [],
        }

    if name.endswith("[default]"):
        controller["is_default"] = True
        name = name.replace("[default]", "")
    elif name.endswith("(public)"):
        controller["is_public"] = True
        name = name.replace("(public)", "")
    elif name.endswith("(random)"):
        controller["is_random"] = True
        name = name.replace("(random)", "")

    controller["name"] = name.strip()

    while next_lines:
        next_line = next_lines.pop()
        result = re.match(_controller_line_pattern, next_line)

        if not result:
            next_lines.append(next_line)
            return controller

        matches = result.groupdict()

        if matches["manufacturer"]:
            controller["manufacturer"] = matches["manufacturer"]
        elif matches["version"]:
            controller["version"] = matches["version"]
        elif matches["name"]:
            controller["name"] = matches["name"]
        elif matches["alias"]:
            controller["alias"] = matches["alias"]
        elif matches["class"]:
            controller["class"] = matches["class"]
        elif matches["powered"]:
            controller["powered"] = matches["powered"]
        elif matches["power_state"]:
            controller["power_state"] = matches["power_state"]
        elif matches["discoverable"]:
            controller["discoverable"] = matches["discoverable"]
        elif matches["discoverable_timeout"]:
            controller["discoverable_timeout"] = matches["discoverable_timeout"]
        elif matches["pairable"]:
            controller["pairable"] = matches["pairable"]
        elif matches["modalias"]:
            controller["modalias"] = matches["modalias"]
        elif matches["discovering"]:
            controller["discovering"] = matches["discovering"]
        elif matches["uuid"]:
            if not "uuids" in controller:
                controller["uuids"] = []
            controller["uuids"].append(matches["uuid"])

    return controller

_device_head_pattern = r"Device (?P<address>([0-9A-F]{2}:){5}[0-9A-F]{2}) (?P<name>.+)"

_device_line_pattern = (
    r"(\s*Name:\s*(?P<name>.+)"
    + r"|\s*Alias:\s*(?P<alias>.+)"
    + r"|\s*Appearance:\s*(?P<appearance>.+)"
    + r"|\s*Class:\s*(?P<class>.+)"
    + r"|\s*Icon:\s*(?P<icon>.+)"
    + r"|\s*Paired:\s*(?P<paired>.+)"
    + r"|\s*Bonded:\s*(?P<bonded>.+)"
    + r"|\s*Trusted:\s*(?P<trusted>.+)"
    + r"|\s*Blocked:\s*(?P<blocked>.+)"
    + r"|\s*Connected:\s*(?P<connected>.+)"
    + r"|\s*LegacyPairing:\s*(?P<legacy_pairing>.+)"
    + r"|\s*Modalias:\s*(?P<modalias>.+)"
    + r"|\s*RSSI:\s*(?P<rssi>.+)"
    + r"|\s*TxPower:\s*(?P<txpower>.+)"
    + r"|\s*Battery\sPercentage:\s*0[xX][0-9a-fA-F]*\s*\((?P<battery_percentage>[0-9]+)\)"
    + r"|\s*UUID:\s*(?P<uuid>.+))"
    + r"|\s*CablePairing:\s*(?P<cable_pairing>.+)"
)


def _parse_device(next_lines: List[str], quiet: bool) -> Optional[Device]:
    next_line = next_lines.pop()
    result = re.match(_device_head_pattern, next_line)

    if not result:
        next_lines.append(next_line)
        return None

    matches = result.groupdict()

    name = matches["name"]

    if name.endswith("not available"):
        return None

    device: Device = {
        "name": '',
        "is_public": False,
        "is_random": False,
        "address": matches["address"],
        "alias": '',
        "appearance": '',
        "class": '',
        "icon": '',
        "paired": '',
        "bonded": '',
        "trusted": '',
        "blocked": '',
        "connected": '',
        "legacy_pairing": '',
        "cable_pairing": '',
        "rssi": 0,
        "txpower": 0,
        "uuids": [],
        "modalias": '',
        "battery_percentage": 0
    }

    if name.endswith("(public)"):
        device["is_public"] = True
        name = name.replace("(public)", "")
    elif name.endswith("(random)"):
        device["is_random"] = True
        name = name.replace("(random)", "")

    device["name"] = name.strip()

    while next_lines:
        next_line = next_lines.pop()
        result = re.match(_device_line_pattern, next_line)

        if not result:
            next_lines.append(next_line)
            return device

        matches = result.groupdict()

        if matches["name"]:
            device["name"] = matches["name"]
        elif matches["alias"]:
            device["alias"] = matches["alias"]
        elif matches["appearance"]:
            device["appearance"] = matches["appearance"]
        elif matches["class"]:
            device["class"] = matches["class"]
        elif matches["icon"]:
            device["icon"] = matches["icon"]
        elif matches["paired"]:
            device["paired"] = matches["paired"]
        elif matches["bonded"]:
            device["bonded"] = matches["bonded"]
        elif matches["trusted"]:
            device["trusted"] = matches["trusted"]
        elif matches["blocked"]:
            device["blocked"] = matches["blocked"]
        elif matches["connected"]:
            device["connected"] = matches["connected"]
        elif matches["legacy_pairing"]:
            device["legacy_pairing"] = matches["legacy_pairing"]
        elif matches["cable_pairing"]:
            device["cable_pairing"] = matches["cable_pairing"]
        elif matches["rssi"]:
            rssi = matches["rssi"]
            try:
                device["rssi"] = int(rssi)
            except ValueError:
                if not quiet:
                    jc.utils.warning_message([f"{next_line} : rssi - {rssi} is not int-able"])
        elif matches["txpower"]:
            txpower = matches["txpower"]
            try:
                device["txpower"] = int(txpower)
            except ValueError:
                if not quiet:
                    jc.utils.warning_message([f"{next_line} : txpower - {txpower} is not int-able"])
        elif matches["uuid"]:
            if not "uuids" in device:
                device["uuids"] = []
            device["uuids"].append(matches["uuid"])
        elif matches["modalias"]:
            device["modalias"] = matches["modalias"]
        elif matches["battery_percentage"]:
            battery_percentage = matches["battery_percentage"]
            try:
                device["battery_percentage"] = int(battery_percentage)
            except ValueError:
                if not quiet:
                    jc.utils.warning_message([f"{next_line} : battery_percentage - {battery_percentage} is not int-able"])

    return device

def parse(data: str, raw: bool = False, quiet: bool = False) -> List[JSONDictType]:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)
    result: List = []

    if jc.utils.has_data(data):
        linedata = data.splitlines()
        linedata.reverse()

        while linedata:
            element = None
            if data.startswith("Controller"):
                element = _parse_controller(linedata)
            elif data.startswith("Device"):
                element = _parse_device(linedata, quiet)  # type: ignore

            if element:
                result.append(element)
            else:
                break

    return result



r"""jc - JSON Convert `rpm -qi` command output parser

Works with `rpm -qi [package]` or `rpm -qia`.

The `..._epoch` calculated timestamp fields are naive. (i.e. based on the
local time of the system the parser is run on)

The `..._epoch_utc` calculated timestamp fields are timezone-aware and are
only available if the timezone field is UTC.

Usage (cli):

    $ rpm -qia | jc --rpm-qi

or

    $ jc rpm -qia

Usage (module):

    import jc
    result = jc.parse('rpm_qi', rpm_qi_command_output)

Schema:

    [
      {
        "name":                     string,
        "epoch":                    integer,
        "version":                  string,
        "release":                  string,
        "architecture":             string,
        "install_date":             string,
        "install_date_epoch":       integer,      # [0]
        "install_date_epoch_utc":   integer,      # [1]
        "group":                    string,
        "size":                     integer,
        "license":                  string,
        "signature":                string,
        "source_rpm":               string,
        "build_date":               string,
        "build_epoch":              integer,      # [0]
        "build_epoch_utc":          integer,      # [1]
        "build_host":               string,
        "relocations":              string,
        "depends": [
                                    string
        ],
        "pre_depends": [
                                    string
        ],
        "recommends": [
                                    string
        ],
        "suggests": [
                                    string
        ],
        "conflicts": [
                                    string
        ],
        "breaks": [
                                    string
        ],
        "tag": [
                                    string
        ],
        "replaces": [
                                    string
        ],
        "packager":                 string,
        "vendor":                   string,
        "url":                      string,
        "summary":                  string,
        "description":              string
      }
    ]

    [0] naive timestamp
    [1] Aware timestamp if timezone is UTC

Examples:

    $ rpm -qia | jc --rpm-qi -p
    [
      {
        "name": "make",
        "epoch": 1,
        "version": "3.82",
        "release": "24.el7",
        "architecture": "x86_64",
        "install_date": "Wed 16 Oct 2019 09:21:42 AM PDT",
        "group": "Development/Tools",
        "size": 1160660,
        "license": "GPLv2+",
        "signature": "RSA/SHA256, Thu 22 Aug 2019 02:34:59 PM PDT, Key ...",
        "source_rpm": "make-3.82-24.el7.src.rpm",
        "build_date": "Thu 08 Aug 2019 05:47:25 PM PDT",
        "build_host": "x86-01.bsys.centos.org",
        "relocations": "(not relocatable)",
        "packager": "CentOS BuildSystem <http://bugs.centos.org>",
        "vendor": "CentOS",
        "url": "http://www.gnu.org/software/make/",
        "summary": "A GNU tool which simplifies the build process for ...",
        "description": "A GNU tool for controlling the generation of ex...",
        "build_epoch": 1565311645,
        "build_epoch_utc": null,
        "install_date_epoch": 1571242902,
        "install_date_epoch_utc": null
      },
      {
        "name": "kbd-legacy",
        "version": "1.15.5",
        "release": "15.el7",
        "architecture": "noarch",
        "install_date": "Thu 15 Aug 2019 10:53:08 AM PDT",
        "group": "System Environment/Base",
        "size": 503608,
        "license": "GPLv2+",
        "signature": "RSA/SHA256, Mon 12 Nov 2018 07:17:49 AM PST, Key ...",
        "source_rpm": "kbd-1.15.5-15.el7.src.rpm",
        "build_date": "Tue 30 Oct 2018 03:40:00 PM PDT",
        "build_host": "x86-01.bsys.centos.org",
        "relocations": "(not relocatable)",
        "packager": "CentOS BuildSystem <http://bugs.centos.org>",
        "vendor": "CentOS",
        "url": "http://ftp.altlinux.org/pub/people/legion/kbd",
        "summary": "Legacy data for kbd package",
        "description": "The kbd-legacy package contains original keymap...",
        "build_epoch": 1540939200,
        "build_epoch_utc": null,
        "install_date_epoch": 1565891588,
        "install_date_epoch_utc": null
      },
      ...
    ]

    $ rpm -qia | jc --rpm-qi -p -r
    [
      {
        "name": "make",
        "epoch": "1",
        "version": "3.82",
        "release": "24.el7",
        "architecture": "x86_64",
        "install_date": "Wed 16 Oct 2019 09:21:42 AM PDT",
        "group": "Development/Tools",
        "size": "1160660",
        "license": "GPLv2+",
        "signature": "RSA/SHA256, Thu 22 Aug 2019 02:34:59 PM PDT, Key ...",
        "source_rpm": "make-3.82-24.el7.src.rpm",
        "build_date": "Thu 08 Aug 2019 05:47:25 PM PDT",
        "build_host": "x86-01.bsys.centos.org",
        "relocations": "(not relocatable)",
        "packager": "CentOS BuildSystem <http://bugs.centos.org>",
        "vendor": "CentOS",
        "url": "http://www.gnu.org/software/make/",
        "summary": "A GNU tool which simplifies the build process for...",
        "description": "A GNU tool for controlling the generation of exe..."
      },
      {
        "name": "kbd-legacy",
        "version": "1.15.5",
        "release": "15.el7",
        "architecture": "noarch",
        "install_date": "Thu 15 Aug 2019 10:53:08 AM PDT",
        "group": "System Environment/Base",
        "size": "503608",
        "license": "GPLv2+",
        "signature": "RSA/SHA256, Mon 12 Nov 2018 07:17:49 AM PST, Key ...",
        "source_rpm": "kbd-1.15.5-15.el7.src.rpm",
        "build_date": "Tue 30 Oct 2018 03:40:00 PM PDT",
        "build_host": "x86-01.bsys.centos.org",
        "relocations": "(not relocatable)",
        "packager": "CentOS BuildSystem <http://bugs.centos.org>",
        "vendor": "CentOS",
        "url": "http://ftp.altlinux.org/pub/people/legion/kbd",
        "summary": "Legacy data for kbd package",
        "description": "The kbd-legacy package contains original keymaps..."
      },
      ...
    ]
"""
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.9'
    description = '`rpm -qi` command parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux']
    magic_commands = ['rpm -qi', 'rpm -qia', 'rpm -qai']
    tags = ['command']


__version__ = info.version


def _process(proc_data):
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        List of Dictionaries. Structured data to conform to the schema.
    """
    int_list = {'epoch', 'size', 'installed_size'}
    split_list = {'depends', 'pre_depends', 'recommends', 'suggests', 'conflicts',
                  'breaks', 'tag', 'replaces'}

    for entry in proc_data:
        for key in entry:
            if key in int_list:
                entry[key] = jc.utils.convert_to_int(entry[key])

            # for apt-cache show output
            if key in split_list:
                val_list = entry[key].split(',')
                val_list = [x.strip() for x in val_list if x]
                entry[key] = val_list

        if 'build_date' in entry:
            timestamp = jc.utils.timestamp(entry['build_date'], format_hint=(3000,))
            entry['build_epoch'] = timestamp.naive
            entry['build_epoch_utc'] = timestamp.utc

        if 'install_date' in entry:
            timestamp = jc.utils.timestamp(entry['install_date'], format_hint=(3000,))
            entry['install_date_epoch'] = timestamp.naive
            entry['install_date_epoch_utc'] = timestamp.utc

    return proc_data


def parse(data, raw=False, quiet=False):
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output = []
    entry_obj = {}
    desc_entry = False
    desc_en_entry = False
    description = []

    if jc.utils.has_data(data):

        for line in filter(None, data.splitlines()):
            split_line = line.split(': ', maxsplit=1)

            if (split_line[0].startswith('Name') or split_line[0] == 'Package') and len(split_line) == 2:
                if entry_obj:
                    if description:
                        entry_obj['description'] = ' '.join(description)
                    raw_output.append(entry_obj)
                    entry_obj = {}
                    desc_entry = False
                    desc_en_entry = False

            if line.startswith('Description :'):
                desc_entry = True
                description = []
                continue

            # for apt-cache show output
            if line.startswith('Description-en:'):
                desc_en_entry = True
                description = [split_line[1].strip()]
                continue

            if desc_entry:
                description.append(line)
                continue

            if desc_en_entry and line.startswith(' '):
                description.append(line)
                continue

            if len(split_line) == 2:
                keyname = jc.utils.normalize_key(split_line[0])
                entry_obj[keyname] = split_line[1].strip()

        if entry_obj:
            if description:
                entry_obj['description'] = ' '.join(description)
            raw_output.append(entry_obj)

    return raw_output if raw else _process(raw_output)



r"""jc - JSON Convert `airport -I` command output parser

The `airport` program can be found at `/System/Library/PrivateFrameworks/Apple80211.framework/Versions/Current/Resources/airport`.

Usage (cli):

    $ airport -I | jc --airport

or

    $ jc airport -I

Usage (module):

    import jc
    result = jc.parse('airport', airport_command_output)

Schema:

    {
      "agrctlrssi":        integer,
      "agrextrssi":        integer,
      "agrctlnoise":       integer,
      "agrextnoise":       integer,
      "state":             string,
      "op_mode":           string,
      "lasttxrate":        integer,
      "maxrate":           integer,
      "lastassocstatus":   integer,
      "802_11_auth":       string,
      "link_auth":         string,
      "bssid":             string,
      "ssid":              string,
      "mcs":               integer,
      "channel":           string
    }

Examples:

    $ airport -I | jc --airport -p
    {
      "agrctlrssi": -66,
      "agrextrssi": 0,
      "agrctlnoise": -90,
      "agrextnoise": 0,
      "state": "running",
      "op_mode": "station",
      "lasttxrate": 195,
      "maxrate": 867,
      "lastassocstatus": 0,
      "802_11_auth": "open",
      "link_auth": "wpa2-psk",
      "bssid": "3c:37:86:15:ad:f9",
      "ssid": "SnazzleDazzle",
      "mcs": 0,
      "channel": "48,80"
    }

    $ airport -I | jc --airport -p -r
    {
      "agrctlrssi": "-66",
      "agrextrssi": "0",
      "agrctlnoise": "-90",
      "agrextnoise": "0",
      "state": "running",
      "op_mode": "station",
      "lasttxrate": "195",
      "maxrate": "867",
      "lastassocstatus": "0",
      "802_11_auth": "open",
      "link_auth": "wpa2-psk",
      "bssid": "3c:37:86:15:ad:f9",
      "ssid": "SnazzleDazzle",
      "mcs": "0",
      "channel": "48,80"
    }
"""
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.5'
    description = '`airport -I` command parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['darwin']
    magic_commands = ['airport -I']
    tags = ['command']


__version__ = info.version


def _process(proc_data):
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (Dictionary) raw structured data to process

    Returns:

        Dictionary. Structured data to conform to the schema.
    """
    int_list = {'agrctlrssi', 'agrextrssi', 'agrctlnoise', 'agrextnoise',
                'lasttxrate', 'maxrate', 'lastassocstatus', 'mcs'}

    for key in proc_data:
        if key in int_list:
            proc_data[key] = jc.utils.convert_to_int(proc_data[key])

    return proc_data


def parse(data, raw=False, quiet=False):
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        Dictionary. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output = {}

    if jc.utils.has_data(data):

        for line in filter(None, data.splitlines()):
            linedata = line.split(':', maxsplit=1)
            key = linedata[0].strip().lower().replace(' ', '_').replace('.', '_')
            value = linedata[1].strip()
            raw_output[key] = value

    if raw:
        return raw_output
    else:
        return _process(raw_output)
